# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku dataset provider."""

import re
import shutil
import urllib
from collections import defaultdict
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, List, Optional

from renku.core import errors
from renku.core.dataset.datasets_provenance import DatasetsProvenance
from renku.core.dataset.providers.api import ImporterApi, ImportProviderInterface, ProviderApi, ProviderPriority
from renku.core.login import read_renku_token
from renku.core.storage import pull_paths_from_storage
from renku.core.util import communication
from renku.core.util.git import clone_renku_repository, get_cache_directory_for_repository, get_file_size
from renku.core.util.metadata import is_external_file, make_project_temp_dir
from renku.core.util.urls import remove_credentials
from renku.domain_model.project_context import project_context

if TYPE_CHECKING:
    from renku.core.dataset.providers.models import DatasetAddMetadata, ProviderDataset, ProviderParameter
    from renku.domain_model.dataset import Dataset


class RenkuProvider(ProviderApi, ImportProviderInterface):
    """Renku API provider."""

    priority = ProviderPriority.HIGH
    name = "Renku"
    is_remote = True

    def __init__(self, uri: str, **_):
        super().__init__(uri=uri)

        self._accept = "application/json"
        self._authorization_header: Optional[Dict[str, str]] = None
        self._gitlab_token: Optional[str] = None
        self._renku_token: Optional[str] = None
        self._tag: Optional[str] = None

    @staticmethod
    def supports(uri):
        """Whether this provider supports a given URI."""
        parsed_url = urllib.parse.urlparse(uri)

        if not parsed_url.netloc:
            return False

        _, dataset_id = RenkuProvider._extract_project_and_dataset_ids(parsed_url)
        return dataset_id is not None

    @staticmethod
    def get_import_parameters() -> List["ProviderParameter"]:
        """Returns parameters that can be set for import."""
        from renku.core.dataset.providers.models import ProviderParameter

        return [ProviderParameter("tag", help="Import a specific tag instead of the latest version.", type=str)]

    def get_importer(self, tag: Optional[str] = None, gitlab_token: Optional[str] = None, **kwargs):
        """Retrieves a dataset import manager from Renku.

        Args:
            tag(Optional[str]): Dataset version to import.
            gitlab_token(Optional[str]): Gitlab access token.

        Returns:
            RenkuImporter: A Renku import manager.
        """
        self._tag = tag
        self._gitlab_token = gitlab_token

        self._prepare_auth(self.uri)

        slug, identifier, latest_version_uri, kg_url = self._fetch_dataset_info(self.uri)

        project_url_ssh, project_url_http = self._get_project_urls(kg_url)

        return RenkuImporter(
            uri=self.uri,
            slug=slug,
            identifier=identifier,
            tag=self._tag,
            latest_version_uri=latest_version_uri,
            project_url_ssh=project_url_ssh,
            project_url_http=project_url_http,
            gitlab_token=self._gitlab_token,
            renku_token=self._renku_token,
        )

    def _fetch_dataset_info(self, uri):
        """Return initial dataset identifier and urls of all projects that contain the dataset."""
        parsed_url = urllib.parse.urlparse(uri)

        project_id, dataset_slug_or_id = RenkuProvider._extract_project_and_dataset_ids(parsed_url)
        if not project_id and not dataset_slug_or_id:
            raise errors.ParameterError("Invalid URI", param_hint=uri)

        kg_path = f"/knowledge-graph/datasets/{dataset_slug_or_id}"
        dataset_kg_url = parsed_url._replace(path=kg_path).geturl()

        try:
            dataset_info = self._query_knowledge_graph(dataset_kg_url)
        except errors.NotFound:
            # NOTE: If URI is not found we assume that it contains dataset's slug instead of its id
            dataset_slug = dataset_slug_or_id
            identifier = None
            dataset_info = None
        else:
            # name was renamed to slug, name kept for backwards compatibility
            dataset_slug = dataset_info.get("slug", dataset_info.get("name"))
            identifier = dataset_info["identifier"]

        if project_id:
            kg_path = f"/knowledge-graph/projects/{project_id}"
            project_kg_url = parsed_url._replace(path=kg_path).geturl()
        elif not dataset_info:
            raise errors.NotFound(
                f"Resource not found in knowledge graph: {uri}\n"
                f"Hint: If the project is private you need to 'renku login {parsed_url.netloc}'"
            )
        else:
            project = dataset_info.get("project", {})
            links = project.get("_links", [])
            project_kg_url = next((link["href"] for link in links if link["rel"] == "project-details"), None)

            if not project_kg_url:
                raise errors.ParameterError("Cannot find project's KG URL from URI", param_hint=uri)

        latest_identifier, latest_version_uri = self._fetch_dataset_info_from_project(project_kg_url, dataset_slug)
        identifier = identifier or latest_identifier

        return dataset_slug, identifier, latest_version_uri, project_kg_url

    def _fetch_dataset_info_from_project(self, project_kg_url, dataset_slug):
        datasets_kg_url = f"{project_kg_url}/datasets"
        try:
            response = self._query_knowledge_graph(datasets_kg_url)
        except errors.NotFound:
            raise errors.NotFound(f"Cannot find project in the knowledge graph: {project_kg_url}")

        dataset = next((d for d in response if d.get("slug") == dataset_slug), None)
        if not dataset:
            raise errors.OperationError(f"Cannot fetch dataset with slug '{dataset_slug}' from '{project_kg_url}'")

        links = dataset.get("_links", [])
        latest_version_uri = next((link["href"] for link in links if link["rel"] == "details"), None)
        if latest_version_uri:
            latest_version_uri = latest_version_uri.replace("/knowledge-graph", "")

        return dataset["identifier"], latest_version_uri

    @staticmethod
    def _extract_project_and_dataset_ids(parsed_url):
        # https://<host>/projects/:namespace/:0-or-more-subgroups/:name/datasets/:dataset-slug
        # https://<host>/projects/:namespace/:0-or-more-subgroups/:name/datasets/:id
        # https://<host>/datasets/:id
        match = re.match(r"(?:/projects/((?:[^/]+/)+[^/]+))?/datasets/([^/]+)/?$", parsed_url.path)
        project_id, dataset_slug_or_id = match.groups() if match else (None, None)
        return project_id, dataset_slug_or_id

    def _query_knowledge_graph(self, url):
        from renku.core.util import requests

        if self._renku_token and not self._gitlab_token:
            # NOTE: Authorization with renku token requires going through the gateway route
            url = url.replace("/knowledge-graph/", "/api/kg/")

        try:
            response = requests.get(url, headers=self._authorization_header)
        except errors.RequestError as e:
            raise errors.OperationError(f"Cannot access knowledge graph: {url}") from e

        parsed_uri = urllib.parse.urlparse(self.uri)
        if response.status_code == 404:
            raise errors.NotFound(
                f"Resource not found in knowledge graph: {url}\n"
                f"Hint: If the project is private you need to 'renku login {parsed_uri.netloc}'"
            )
        elif response.status_code in [401, 403]:
            raise errors.OperationError(
                f"Unauthorized access to knowledge graph: Run 'renku login {parsed_uri.netloc}'"
            )
        elif response.status_code != 200:
            raise errors.OperationError(f"Cannot access knowledge graph: {url}\nResponse code: {response.status_code}")

        return response.json()

    def _get_project_urls(self, project_kg_url):
        json = self._query_knowledge_graph(project_kg_url)
        urls = json.get("urls", {})

        return urls.get("ssh"), urls.get("http")

    def _prepare_auth(self, uri):
        if self._gitlab_token:
            token = self._gitlab_token
        else:
            self._renku_token = read_renku_token(endpoint=uri)
            token = self._renku_token

        self._authorization_header = {"Authorization": f"Bearer {token}"} if token else {}


class RenkuImporter(ImporterApi):
    """Renku record serializer."""

    def __init__(
        self,
        uri,
        slug,
        identifier,
        tag,
        latest_version_uri,
        project_url_ssh,
        project_url_http,
        gitlab_token,
        renku_token,
    ):
        """Create a RenkuImporter from a Dataset."""
        super().__init__(uri=uri, original_uri=uri)

        self._slug = slug
        self._identifier = identifier
        self._tag = tag
        self._latest_version_uri = latest_version_uri
        self._project_url_ssh = project_url_ssh
        self._project_url_http = project_url_http
        self._gitlab_token = gitlab_token
        self._renku_token = renku_token

        self._project_url = None
        self._remote_repository = None
        self._remote_path: Optional[Path] = None

    def fetch_provider_dataset(self) -> "ProviderDataset":
        """Return encapsulated dataset instance."""
        self._provider_dataset = self._fetch_dataset()
        assert self._provider_dataset is not None, "Dataset wasn't fetched."
        return self._provider_dataset

    def download_files(self, destination: Path, extract: bool) -> List["DatasetAddMetadata"]:
        """Download dataset files from the remote provider."""
        from renku.core.dataset.providers.models import DatasetAddAction, DatasetAddMetadata
        from renku.domain_model.dataset import RemoteEntity

        url = remove_credentials(self.project_url)

        dataset_datadir = self.provider_dataset.get_datadir()
        remote_repository = self.repository

        if self.provider_dataset.version:  # NOTE: A tag was specified for import
            sources, checksums = zip(*[(f.path, f.checksum) for f in self.provider_dataset_files])  # type: ignore
        else:
            sources = [f.path for f in self.provider_dataset_files]  # type: ignore
            checksums = None

        assert destination.exists() and destination.is_dir(), "Destination dir must exist when importing a dataset"
        if checksums is not None:
            assert len(checksums) == len(sources), "Each source must have a corresponding checksum"  # type: ignore

        def add_file(src_entity_path: str, content_path: Path, checksum) -> None:
            """
            Create a DatasetAddMetadata.

            Args:
                src_entity_path: Entity path from the source dataset which is a relative path.
                content_path: Absolute path of the file content when copied with a checksum.
                checksum: Entity checksum.
            """
            try:
                relative_path = Path(src_entity_path).relative_to(dataset_datadir)
            except ValueError:  # Files that are not in dataset's data directory
                relative_path = Path(src_entity_path)

            dst = destination / relative_path
            path_in_dst_repo = dst.relative_to(project_context.path)

            already_copied = path_in_dst_repo in new_files  # A path with the same destination is already copied
            new_files[path_in_dst_repo].append(src_entity_path)
            if already_copied:
                return

            if is_external_file(path=src_entity_path, project_path=remote_repository.path):
                source = (remote_repository.path / src_entity_path).resolve()
                action = DatasetAddAction.SYMLINK
            else:
                source = content_path
                action = DatasetAddAction.MOVE

            checksum = checksum or remote_repository.get_object_hash(  # type: ignore
                revision="HEAD", path=src_entity_path
            )
            if not checksum:
                raise errors.FileNotFound(f"Cannot find '{file}' in the remote project")

            new_file = DatasetAddMetadata(
                entity_path=path_in_dst_repo,
                url=url,
                based_on=RemoteEntity(checksum=checksum, path=src_entity_path, url=url),
                action=action,
                source=source,
                destination=dst,
            )
            results.append(new_file)

        results: List["DatasetAddMetadata"] = []
        new_files: Dict[Path, List[str]] = defaultdict(list)

        if self.provider_dataset.storage:  # NOTE: Dataset with a backend storage
            results = [
                DatasetAddMetadata(
                    entity_path=Path(f.path),
                    url=url,
                    based_on=RemoteEntity(checksum=f.checksum, path=f.path, url=url),
                    action=DatasetAddAction.METADATA_ONLY,
                    source=remote_repository.path / f.path,
                    destination=destination / f.path,
                )
                for f in self.provider_dataset_files
            ]
        elif checksums is None:
            with project_context.with_path(remote_repository.path):
                pull_paths_from_storage(
                    project_context.repository, *(remote_repository.path / p for p in sources)  # type: ignore
                )

            for file in sources:
                add_file(file, content_path=remote_repository.path / file, checksum=None)  # type: ignore
        else:  # NOTE: Renku dataset import with a tag
            content_path_root = make_project_temp_dir(project_context.path)
            content_path_root.mkdir(parents=True, exist_ok=True)
            filename = 1

            for file, checksum in zip(sources, checksums):  # type: ignore
                content_path = content_path_root / str(filename)
                filename += 1

                try:
                    remote_repository.copy_content_to_file(  # type: ignore
                        path=file, checksum=checksum, output_path=content_path
                    )
                except errors.FileNotFound:
                    raise errors.FileNotFound(f"Cannot find '{file}' with hash '{checksum}' in the remote project")

                add_file(file, content_path=content_path, checksum=checksum)

        duplicates = [v for v in new_files.values() if len(v) > 1]
        if duplicates:
            files = {str(p) for paths in duplicates for p in paths}
            files_str = "/n/t".join(sorted(files))
            communication.warn(f"The following files overwrite each other in the destination project:/n/t{files_str}")

        return results

    def tag_dataset(self, slug: str) -> None:
        """Create a tag for the dataset ``slug`` if the remote dataset has a tag/version."""
        from renku.core.dataset.tag import add_dataset_tag

        if self.provider_dataset.tag:
            add_dataset_tag(
                dataset_slug=slug,
                tag=self.provider_dataset.tag.name,
                description=self.provider_dataset.tag.description,
            )
        elif self.provider_dataset.version:
            add_dataset_tag(
                dataset_slug=slug,
                tag=self.provider_dataset.version,
                description=f"Tag {self.provider_dataset.version} created by renku import",
            )

    def copy_extra_metadata(self, new_dataset: "Dataset") -> None:
        """Copy provider specific metadata once the dataset is created."""
        if not self.provider_dataset.images:
            return

        for image in self.provider_dataset.images:
            if image.is_absolute:
                continue

            assert isinstance(self._remote_path, Path)  # TODO: Replace with proper typing

            remote_image_path = self._remote_path / image.content_url
            local_image_path = project_context.path / image.content_url
            local_image_path.parent.mkdir(exist_ok=True, parents=True)

            shutil.copy(remote_image_path, local_image_path)

        new_dataset.images = self.provider_dataset.images or []

    def is_latest_version(self):
        """Check if dataset is at last possible version."""
        return self.latest_uri.endswith(self._identifier)

    def is_version_equal_to(self, dataset: Any) -> bool:
        """Check if a dataset has the identifier as the record."""
        same_as = getattr(dataset, "same_as", None)
        return same_as is not None and same_as.value.endswith(self._identifier)

    @property
    def project_url(self):
        """URL of the Renku project in Gitlab."""
        return self._project_url

    @property
    def repository(self):
        """The cloned repository that contains the dataset."""
        return self._remote_repository

    @staticmethod
    def _extract_dataset_id(uri):
        """Extract dataset id from uri."""
        u = urllib.parse.urlparse(uri)
        return Path(u.path).name

    @property
    def version(self):
        """Get record version."""
        return self._provider_dataset.version if self._provider_dataset else None

    @property
    def latest_uri(self):
        """Get URI of the latest version."""
        return self._latest_version_uri

    @property
    def datadir_exists(self):
        """Whether the dataset data directory exists (might be missing in git if empty)."""
        if self._remote_path is None:
            raise errors.DatasetImportError("Dataset not fetched.")
        return (self._remote_path / self.provider_dataset.get_datadir()).exists()

    def _fetch_dataset(self):
        from renku.core.dataset.providers.models import ProviderDataset, ProviderDatasetFile
        from renku.domain_model.dataset import Url

        remote_repository = None

        parsed_uri = urllib.parse.urlparse(self.uri)

        urls = (self._project_url_ssh, self._project_url_http)
        # Clone the project
        communication.echo(msg="Cloning remote repository...")
        for url in urls:
            try:
                remote_repository = clone_renku_repository(
                    url=url,
                    path=get_cache_directory_for_repository(url=url),
                    gitlab_token=self._gitlab_token,
                    deployment_hostname=parsed_uri.netloc,
                    depth=None,
                    reuse_existing_repository=True,
                    use_renku_credentials=True,
                )
            except errors.GitError:
                pass
            else:
                self._project_url = url
                break

        if self._project_url is None or remote_repository is None:
            raise errors.ParameterError("Cannot clone remote projects:\n\t" + "\n\t".join(urls), param_hint=self.uri)

        with project_context.with_path(remote_repository.path):
            self._remote_path = project_context.path

            self._migrate_project()
            self._remote_repository = remote_repository

            datasets_provenance = DatasetsProvenance()

            dataset = datasets_provenance.get_by_slug(self._slug)
            if not dataset:
                raise errors.ParameterError(f"Cannot find dataset '{self._slug}' in project '{self._project_url}'")

            if self._tag:
                tags = datasets_provenance.get_all_tags(dataset=dataset)
                tag = next((t for t in tags if t.name == self._tag), None)

                if tag is None:
                    raise errors.ParameterError(f"Cannot find tag '{self._tag}' for dataset '{self._slug}'")

                dataset = datasets_provenance.get_by_id(tag.dataset_id.value)
            else:
                tag = None

            assert dataset is not None
            provider_dataset = ProviderDataset.from_dataset(dataset)

            # NOTE: Set the dataset version to the given tag (to reset the version if no tag was provided)
            provider_dataset.version = self._tag
            # NOTE: Store the tag so that it can be checked later to see if a tag was specified for import
            provider_dataset.tag = tag

        provider_dataset.derived_from = None
        provider_dataset.same_as = Url(url_id=remove_credentials(self.latest_uri))

        self._provider_dataset_files = [
            ProviderDatasetFile(
                path=file.entity.path,
                checksum=file.entity.checksum,
                filename=Path(file.entity.path).name,
                filetype=Path(file.entity.path).suffix.replace(".", ""),
                filesize=get_file_size(self._remote_path, file.entity.path),
                source=file.source,
            )
            for file in dataset.files
        ]

        return provider_dataset

    @staticmethod
    def _migrate_project():
        from renku.core.migration.migrate import is_project_unsupported, migrate_project  # Slow import
        from renku.core.migration.models.migration import MigrationType

        if is_project_unsupported():
            return

        try:
            communication.disable()
            # NOTE: We are not interested in migrating workflows when importing datasets
            migrate_project(
                skip_template_update=True, skip_docker_update=True, migration_type=~MigrationType.WORKFLOWS, strict=True
            )
        finally:
            communication.enable()



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Represent the Common Workflow Language types."""

import os
from pathlib import Path
from typing import Any, List, Optional


class _PathFormatterMixin:
    """Format path property."""

    path: Path

    def __str__(self):
        """Simple conversion to string."""
        return os.path.relpath(os.path.abspath(str(self.path)), os.path.realpath(os.getcwd()))


class File(_PathFormatterMixin):
    """Represent a file."""

    def __init__(self, path: Path, mime_type: Optional[List[str]] = None) -> None:
        self.path = path
        self.mime_type = mime_type or ["application/octet-stream"]


class Directory(_PathFormatterMixin):
    """Represent a directory."""

    def __init__(self, path: Path, listing: Optional[List[Any]] = None) -> None:
        self.path = path  # TODO add validation to allow only directories
        self.listing = listing or []


PATH_OBJECTS = {"File", "Directory"}
PATH_TYPES = (File, Directory)



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Dataset jobs."""
import os
import urllib

from urllib3.exceptions import HTTPError

from renku.command.dataset import add_to_dataset_command, import_dataset_command
from renku.core import errors
from renku.core.util.contexts import renku_project_context
from renku.core.util.git import push_changes
from renku.domain_model.git import GitURL
from renku.infrastructure.repository import Repository
from renku.ui.service.logger import worker_log
from renku.ui.service.utils.callback import ServiceCallback
from renku.ui.service.views.decorators import requires_cache


@requires_cache
def dataset_import(
    cache,
    user,
    user_job_id,
    project_id,
    dataset_uri,
    slug=None,
    extract=False,
    tag=None,
    timeout=None,
    commit_message=None,
    data_directory=None,
):
    """Job for dataset import."""
    user = cache.ensure_user(user)
    worker_log.debug(f"executing dataset import job for {user.user_id}:{user.fullname}")

    user_job = cache.get_job(user, user_job_id)
    user_job.in_progress()

    try:
        worker_log.debug(f"retrieving metadata for project {project_id}")
        project = cache.get_project(user, project_id)
        with renku_project_context(project.abs_path):
            worker_log.debug(f"project found in cache - importing dataset {dataset_uri}")
            communicator = ServiceCallback(user_job=user_job)

            gitlab_token = user.token if _is_safe_to_pass_gitlab_token(project.git_url, dataset_uri) else None

            command = import_dataset_command().with_commit_message(commit_message)
            command.with_communicator(communicator).build().execute(
                uri=dataset_uri,
                slug=slug,
                extract=extract,
                tag=tag,
                yes=True,
                gitlab_token=gitlab_token,
                datadir=data_directory,
            )

            worker_log.debug("operation successful - syncing with remote")
            remote_branch = push_changes(Repository(project.abs_path), remote="origin")
            user_job.update_extras("remote_branch", remote_branch)

            user_job.complete()
            worker_log.debug("job completed")
    except (HTTPError, errors.ParameterError, errors.RenkuException, errors.GitCommandError) as exp:
        user_job.fail_job(str(exp))

        # Reraise exception, so we see trace in job metadata
        # and in metrics as failed job.
        raise exp


def _is_safe_to_pass_gitlab_token(project_git_url, dataset_uri):
    """Passing token is safe if project and dataset belong to the same deployment."""
    project_host = os.environ.get("RENKU_DOMAIN") or GitURL.parse(project_git_url).hostname
    dataset_host = urllib.parse.urlparse(dataset_uri).netloc

    # NOTE: URLs changed from domain/gitlab to gitlab.domain when moving to cloud native gitlab
    if project_host.startswith("gitlab.") and not dataset_host.startswith("gitlab."):
        project_host = project_host.replace("gitlab.", "", 1)

    return project_host == dataset_host


@requires_cache
def dataset_add_remote_file(cache, user, user_job_id, project_id, create_dataset, commit_message, slug, url):
    """Add a remote file to a specified dataset."""
    user = cache.ensure_user(user)
    worker_log.debug(f"executing dataset add remote file job for {user.user_id}:{user.fullname}")

    user_job = cache.get_job(user, user_job_id)
    user_job.in_progress()

    try:
        worker_log.debug(f"checking metadata for project {project_id}")
        project = cache.get_project(user, project_id)

        with renku_project_context(project.abs_path):
            urls = url if isinstance(url, list) else [url]

            worker_log.debug(f"adding files {urls} to dataset {slug}")
            command = add_to_dataset_command().with_commit_message(commit_message).build()
            result = command.execute(dataset_slug=slug, urls=urls, create=create_dataset)
            if result.error:
                raise result.error

            worker_log.debug("operation successful - syncing with remote")
            remote_branch = push_changes(Repository(project.abs_path), remote="origin")
            user_job.update_extras("remote_branch", remote_branch)

            user_job.complete()
            worker_log.debug("job completed")
    except (HTTPError, BaseException, errors.GitCommandError, errors.RenkuException) as exp:
        user_job.fail_job(str(exp))

        # Reraise exception, so we see trace in job metadata
        # and in metrics as failed job.
        raise exp



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku service version view."""
from flask import Blueprint

from renku.ui.service.config import SERVICE_PREFIX
from renku.ui.service.controllers.versions_list import VersionsListCtrl
from renku.ui.service.views.error_handlers import handle_common_except

VERSIONS_LIST_BLUEPRINT_TAG = "versions"
versions_list_blueprint = Blueprint("versions", __name__, url_prefix=SERVICE_PREFIX)


@versions_list_blueprint.route("/versions", methods=["GET"], provide_automatic_options=False)
@handle_common_except
def versions_list():
    """
    Shows the list of all supported metadata versions.

    ---
    get:
      description: Metadata versions supported by all deployed core services.
      responses:
        200:
          description: The list of metadata versions.
          content:
            application/json:
              schema: VersionsListResponseRPC
      tags:
        - version
    """
    return VersionsListCtrl().to_response()



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku service datasets add controller."""

from renku.ui.service.controllers.datasets_add_file import DatasetsAddFileCtrl
from renku.ui.service.serializers.v1.datasets import DatasetAddRequest_2_1, DatasetAddResponseRPC_2_1


class DatasetsAddFileCtrl_2_1(DatasetsAddFileCtrl):
    """Controller for datasets add endpoint."""

    REQUEST_SERIALIZER = DatasetAddRequest_2_1()  # type: ignore
    RESPONSE_SERIALIZER = DatasetAddResponseRPC_2_1()  # type: ignore



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Migrate datasets based on Git submodules."""

import glob
import os
import shutil
from pathlib import Path

from renku.core import errors
from renku.core.migration.m_0009__new_metadata_storage import fetch_datasets
from renku.core.migration.models.v3 import DatasetFileSchemaV3, get_project_datasets
from renku.core.migration.models.v9 import DatasetFile, OldDatasetFileSchema, generate_file_id, generate_label
from renku.core.util.urls import remove_credentials
from renku.domain_model.project_context import project_context
from renku.infrastructure.repository import Repository


def migrate(migration_context):
    """Migration function."""
    _migrate_submodule_based_datasets(migration_context)


def _migrate_submodule_based_datasets(migration_context):
    from renku.core.migration.migrate import is_project_unsupported, migrate_project

    submodules = project_context.repository.submodules
    if len(submodules) == 0:
        return

    repo_paths = []
    symlinks = []

    submodules.update()

    for dataset in get_project_datasets():
        for file_ in dataset.files:
            path = project_context.path / file_.path
            if not path.is_symlink():
                continue

            target = path.resolve()

            if "/.renku/vendors/" not in str(target):
                continue

            repo_path = Repository(path=target.parent, search_parent_directories=True).path
            if repo_path not in repo_paths:
                repo_paths.append(repo_path)

            symlinks.append((file_.path, target, repo_path))

    if not symlinks:
        return

    submodules_urls = {s.relative_path: s.url for s in submodules}

    for repo_path in repo_paths:
        with project_context.with_path(repo_path, save_changes=True):
            if not is_project_unsupported():
                migrate_project(skip_template_update=True, skip_docker_update=True)

    metadata = {}

    for path, target, repo_path in symlinks:
        path_within_repo = target.relative_to(repo_path)

        repo_is_remote = ".renku/vendors/local" not in str(repo_path)
        based_on = None
        submodule_path = repo_path.relative_to(project_context.path)

        url = submodules_urls.get(submodule_path, "")

        if repo_is_remote:
            with project_context.with_path(repo_path):
                based_on = _fetch_file_metadata(migration_context=migration_context, path=path_within_repo)
                if based_on:
                    based_on.url = url
                    based_on.based_on = None
                else:
                    based_on = DatasetFile.from_revision(path=path_within_repo, url=url)
                data = OldDatasetFileSchema().dump(based_on)
                based_on = DatasetFileSchemaV3().load(data)
        else:
            if url:
                full_path = Path(url) / path_within_repo
                rel_path = os.path.relpath(full_path, project_context.path)
                url = f"file://{rel_path}"

        metadata[path] = (based_on, url)

        path = project_context.path / path
        path.unlink()

        try:
            shutil.move(target, path)
        except FileNotFoundError:
            raise errors.InvalidFileOperation(f"File was not found: {target}")

    repository = project_context.repository

    for submodule in submodules:
        if str(submodule.relative_path).startswith(".renku/vendors/"):
            try:
                repository.submodules.remove(submodule, force=True)
            except errors.GitError:
                pass

    for dataset in get_project_datasets():
        for file_ in dataset.files:
            if file_.path in metadata:
                based_on, url = metadata[file_.path]
                file_.based_on = based_on
                file_.url = remove_credentials(url)
                file_.commit = repository.get_previous_commit(file_.path)
                file_._id = generate_file_id(hexsha=file_.commit.hexsha, path=file_.path)
                file_._label = generate_label(file_.path, file_.commit.hexsha)

        dataset.to_yaml()


def _fetch_file_metadata(migration_context, path):
    """Return metadata for a single file."""
    datasets, _ = fetch_datasets(
        migration_context=migration_context,
        revision=project_context.repository.head.commit.hexsha,
        paths=glob.glob(f"{project_context.path}/.renku/datasets/*/*.yml" ""),
        deleted_paths=[],
    )

    for dataset in datasets:
        for file in dataset.files:
            if file.entity.path == path:
                return file



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku service config."""
import os
import tempfile
from pathlib import Path

try:
    import importlib_resources  # type: ignore[import]
except ImportError:
    import importlib.resources as importlib_resources  # type: ignore

SVC_ERROR_GENERIC = 0
SVC_ERROR_USER = 1000
SVC_ERROR_PROGRAMMING = 2000
SVC_ERROR_INTERMITTENT = 3000

DOCS_URL_BASE = "https://renku-python.readthedocs.io/en/latest/"
DOCS_URL_ERRORS = "service_errors.html#renku.ui.service.errors."
ERROR_NOT_AVAILABLE = "Unknown"

SERVICE_NAME = "Renku Service"
OPENAPI_VERSION = "3.0.3"
API_VERSION = "v1"

PROJECT_CLONE_NO_DEPTH = None
PROJECT_CLONE_DEPTH_DEFAULT = int(os.getenv("PROJECT_CLONE_DEPTH_DEFAULT", 1))
TEMPLATE_CLONE_DEPTH_DEFAULT = int(os.getenv("TEMPLATE_CLONE_DEPTH_DEFAULT", 0))

# NOTE: Defaults to 1GB
MAX_CONTENT_LENGTH = int(os.getenv("MAX_CONTENT_LENGTH", 1073741824))

CACHE_DIR = os.getenv("CACHE_DIR", os.path.realpath(tempfile.TemporaryDirectory().name))
CACHE_UPLOADS_PATH = Path(CACHE_DIR) / Path("uploads")
CACHE_UPLOADS_PATH.mkdir(parents=True, exist_ok=True)

CACHE_PROJECTS_PATH = Path(CACHE_DIR) / Path("projects")
CACHE_PROJECTS_PATH.mkdir(parents=True, exist_ok=True)

TAR_ARCHIVE_CONTENT_TYPE = "application/x-tar"
ZIP_ARCHIVE_CONTENT_TYPE = "application/zip"
GZ_ARCHIVE_CONTENT_TYPE = "application/x-gzip"

SUPPORTED_ARCHIVES = [
    TAR_ARCHIVE_CONTENT_TYPE,
    ZIP_ARCHIVE_CONTENT_TYPE,
    GZ_ARCHIVE_CONTENT_TYPE,
]

# the path prefix on the service
SERVICE_PREFIX = os.getenv("CORE_SERVICE_PREFIX", "/")
# the reverse proxy prefix
SERVICE_API_BASE_PATH = os.getenv("CORE_SERVICE_API_BASE_PATH", "/")
# path to the swagger spec
API_SPEC_URL = SERVICE_PREFIX.lstrip("/") + "/spec.json"
# URL for fetching the OIDC configuration
OIDC_URL = os.getenv("OIDC_URL", "/auth/realms/Renku/.well-known/openid-configuration")

ref = importlib_resources.files("renku.ui.service") / "logging.yaml"
with importlib_resources.as_file(ref) as path:
    LOGGER_CONFIG_FILE = path
MESSAGE_PREFIX = os.environ.get("RENKU_DOMAIN", "service") + ":"

# Sentry configuration
SENTRY_ENABLED = os.getenv("SENTRY_ENABLED", "false").lower() == "true"
SENTRY_SAMPLERATE = float(os.getenv("SENTRY_SAMPLE_RATE", 0.2))

# List of all available metadata versions
METADATA_VERSIONS_LIST = os.getenv("METADATA_VERSIONS_LIST", "/svc/config/metadata-versions/metadata-versions.json")



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""local machine executor provider."""

import contextlib
import itertools
import os
import subprocess
import traceback
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, Generator

import networkx as nx

from renku.core import errors
from renku.core.plugin import hookimpl
from renku.core.plugin.provider import RENKU_ENV_PREFIX
from renku.core.util import communication
from renku.domain_model.workflow.provider import IWorkflowProvider

if TYPE_CHECKING:
    from renku.domain_model.workflow.plan import Plan


class LocalWorkflowProvider(IWorkflowProvider):
    """A workflow executor provider on the local machine."""

    @hookimpl
    def workflow_provider(self):
        """Workflow provider name."""
        return self, "local"

    @hookimpl
    def workflow_execute(self, dag: nx.DiGraph, basedir: Path, config: Dict[str, Any]):
        """Executes a given workflow."""
        for plan in nx.topological_sort(dag):
            execute_plan(plan)


def execute_plan(plan: "Plan"):
    """Execute a plan on the local machine."""
    # NOTE: Quoting string values causes a double quoting when passed to ``subprocess.run``
    command_line = plan.to_argv(quote_string=False)

    os_env = os.environ.copy()
    env = get_workflow_parameters_env_vars(workflow=plan)
    os_env.update(env)

    try:
        command_str = " ".join(plan.to_argv(with_streams=True))
        communication.echo(f"Executing step '{plan.name}': '{command_str}' ...")
        with get_plan_std_stream_mapping(plan) as std_streams_mappings:
            return_code = subprocess.run(command_line, cwd=os.getcwd(), env=os_env, **std_streams_mappings).returncode
    except OSError:
        tb = "\n  ".join(traceback.format_exc().split("\n"))
        raise errors.WorkflowExecuteError(f"Execution of step '{plan.name}' failed:\n\n  {tb}", show_prefix=False)

    success_codes = plan.success_codes or [0]
    if return_code not in success_codes:
        message = f"Execution of step '{plan.name}' returned {return_code} exit status which is not in {success_codes}"
        raise errors.InvalidSuccessCode(return_code=return_code, message=message)


@contextlib.contextmanager
def get_plan_std_stream_mapping(plan: "Plan") -> Generator[Dict[str, Any], None, None]:
    """Return the mapping between std streams and file paths for a plan."""
    arguments = itertools.chain(plan.inputs, plan.outputs)
    std_arguments = [a for a in arguments if a.mapped_to]  # type: ignore

    for argument in std_arguments:
        Path(str(argument.actual_value)).parent.mkdir(parents=True, exist_ok=True)

    open_mode = {"stdin": "r", "stdout": "w", "stderr": "w"}
    streams = {}

    with contextlib.ExitStack() as stack:
        for a in std_arguments:
            file = stack.enter_context(open(str(a.actual_value), open_mode[a.mapped_to.stream_type]))  # type: ignore
            streams[a.mapped_to.stream_type] = file  # type: ignore

        yield streams


def get_workflow_parameters_env_vars(workflow: "Plan") -> Dict[str, str]:
    """Get env vars for all workflow inputs/outputs/parameters."""
    env = {}

    for parameter in itertools.chain(workflow.inputs, workflow.parameters):
        env[f"{RENKU_ENV_PREFIX}{parameter.name}"] = str(parameter.actual_value)

    for o in workflow.outputs:
        env[f"{RENKU_ENV_PREFIX}{o.name}"] = str(o.actual_value)
        output_path = Path(o.actual_value)
        if len(output_path.parts) > 1:
            output_path.parent.mkdir(parents=True, exist_ok=True)

    return env



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Repository datasets management."""

from renku.command.command_builder.command import Command
from renku.core.constant import DATASET_METADATA_PATHS
from renku.core.dataset.dataset import (
    create_dataset,
    edit_dataset,
    export_dataset,
    file_unlink,
    import_dataset,
    list_dataset_files,
    list_datasets,
    mount_cloud_storage,
    pull_cloud_storage,
    remove_dataset,
    search_datasets,
    show_dataset,
    unmount_cloud_storage,
    update_datasets,
)
from renku.core.dataset.dataset_add import add_to_dataset
from renku.core.dataset.tag import add_dataset_tag, list_dataset_tags, remove_dataset_tags


def search_datasets_command():
    """Command to get all the datasets whose slug starts with the given string."""
    return Command().command(search_datasets).require_migration().with_database()


def list_datasets_command():
    """Command for listing datasets."""
    return Command().command(list_datasets).with_database().require_migration()


def create_dataset_command():
    """Return a command for creating an empty dataset in the current repository."""
    command = Command().command(create_dataset).lock_dataset().with_database(write=True)
    return command.require_migration().with_commit(commit_only=DATASET_METADATA_PATHS)


def edit_dataset_command():
    """Command for editing dataset metadata."""
    command = Command().command(edit_dataset).lock_dataset().with_database(write=True)
    return command.require_migration().with_commit(commit_only=DATASET_METADATA_PATHS)


def show_dataset_command():
    """Command for showing detailed dataset information."""
    return Command().command(show_dataset).with_database().require_migration()


def add_to_dataset_command():
    """Create a command for adding data to datasets."""
    command = Command().command(add_to_dataset).lock_dataset().with_database(write=True)
    return command.require_migration().with_commit(raise_if_empty=True, commit_only=DATASET_METADATA_PATHS)


def list_files_command():
    """Command for listing dataset files."""
    return Command().command(list_dataset_files).with_database().require_migration()


def file_unlink_command():
    """Command for removing matching files from a dataset."""
    command = Command().command(file_unlink).lock_dataset().with_database(write=True)
    return command.require_migration().with_commit(commit_only=DATASET_METADATA_PATHS)


def remove_dataset_command():
    """Command for deleting a dataset."""
    command = Command().command(remove_dataset).lock_dataset().with_database(write=True)
    return command.require_migration().with_commit(commit_only=DATASET_METADATA_PATHS)


def export_dataset_command():
    """Command for exporting a dataset to 3rd party provider."""
    command = Command().command(export_dataset).with_database()
    return command.require_migration().require_clean()


def import_dataset_command():
    """Create a command for importing datasets."""
    command = Command().command(import_dataset).lock_dataset().with_database(write=True)
    return command.require_migration().with_commit(commit_only=DATASET_METADATA_PATHS)


def update_datasets_command(dry_run=False):
    """Command for updating datasets."""
    command = Command().command(update_datasets).lock_dataset().with_database(write=not dry_run).require_migration()

    if not dry_run:
        command = command.with_commit(commit_only=DATASET_METADATA_PATHS)

    return command


def add_dataset_tag_command():
    """Command for creating a new tag for a dataset."""
    command = Command().command(add_dataset_tag).lock_dataset().with_database(write=True)
    return command.require_migration().with_commit(commit_only=DATASET_METADATA_PATHS)


def remove_dataset_tags_command():
    """Command for removing tags from a dataset."""
    command = Command().command(remove_dataset_tags).lock_dataset().with_database(write=True)
    return command.require_migration().with_commit(commit_only=DATASET_METADATA_PATHS)


def list_tags_command():
    """Command for listing a dataset's tags."""
    return Command().command(list_dataset_tags).with_database().require_migration()


def pull_cloud_storage_command():
    """Command for pulling/copying data from a cloud storage."""
    command = Command().command(pull_cloud_storage).lock_dataset().with_database(write=True)
    return command.require_migration().with_commit(commit_only=DATASET_METADATA_PATHS)


def mount_cloud_storage_command(unmount: bool):
    """Command for mounting a cloud storage."""
    command = unmount_cloud_storage if unmount else mount_cloud_storage
    return Command().command(command).lock_dataset().with_database(write=False).require_migration()


def unmount_cloud_storage_command():
    """Command for unmounting a cloud storage."""
    return Command().command(unmount_cloud_storage).lock_dataset().with_database(write=False).require_migration()



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Classes for integration with Calamus."""

import copy
import inspect

import marshmallow
from calamus import fields
from calamus.schema import JsonLDSchema as CalamusJsonLDSchema
from calamus.utils import normalize_type, normalize_value
from marshmallow.base import SchemaABC

from renku.core import errors
from renku.domain_model.project_context import project_context

prov = fields.Namespace("http://www.w3.org/ns/prov#")
rdfs = fields.Namespace("http://www.w3.org/2000/01/rdf-schema#")
renku = fields.Namespace("https://swissdatasciencecenter.github.io/renku-ontology#")
schema = fields.Namespace("http://schema.org/")
oa = fields.Namespace("http://www.w3.org/ns/oa#")
dcterms = fields.Namespace("http://purl.org/dc/terms/")


class JsonLDSchema(CalamusJsonLDSchema):
    """Base schema class for Renku."""

    def __init__(self, *args, commit=None, **kwargs):
        """Create an instance."""
        self._commit = commit
        super().__init__(*args, **kwargs)

    def _deserialize(self, *args, **kwargs):
        data = super()._deserialize(*args, **kwargs)
        const_args = inspect.signature(self.opts.model)
        parameters = const_args.parameters.values()

        if any(p.name == "commit" for p in parameters):
            if self._commit:
                self._add_field_to_data(data, "commit", self._commit)
            elif (
                project_context.has_context()
                and "_label" in data
                and data["_label"]
                and "@UNCOMMITTED" not in data["_label"]
                and "@" in data["_label"]
            ):
                try:
                    self._add_field_to_data(
                        data,
                        "commit",
                        project_context.repository.get_commit(data["_label"].rsplit("@", maxsplit=1)[-1]),
                    )
                except errors.GitCommitNotFoundError:
                    # NOTE: This means the commit does not exist in the local repository. Could be an external file?
                    pass

        return data

    def _add_field_to_data(self, data, name, value):
        if value:
            if name in data:
                raise ValueError(f"Field {name} is already in data {data}")
            data[name] = value


class Uri(fields._JsonLDField, marshmallow.fields.String, marshmallow.fields.Dict):
    """A Dict/String field."""

    def __init__(self, *args, **kwargs):
        """Create an instance."""
        super().__init__(*args, **kwargs)

    def _serialize(self, value, attr, obj, **kwargs):
        if isinstance(value, str):
            value = super(fields._JsonLDField, self)._serialize(value, attr, obj, **kwargs)
            if self.parent.opts.add_value_types:
                value = {"@value": value, "@type": "http://www.w3.org/2001/XMLSchema#string"}
        elif isinstance(value, dict):
            value = super(marshmallow.fields.String, self)._serialize(value, attr, obj, **kwargs)

        return value

    def _deserialize(self, value, attr, data, **kwargs):
        value = normalize_value(value)
        if not value:
            return None
        elif isinstance(value, str):
            return value
        elif isinstance(value, dict):
            return super(marshmallow.fields.String, self)._deserialize(value, attr, data, **kwargs)
        else:
            raise ValueError(f"Invalid type for field {self.name}: {type(value)}")


class StringList(fields.String):
    """A String field that might be a list when deserializing."""

    def __init__(self, *args, return_max_value=True, **kwargs):
        """Create an instance."""
        super().__init__(*args, **kwargs)
        self.return_max_value = return_max_value

    def _deserialize(self, value, attr, data, **kwargs):
        value = normalize_value(value)

        if isinstance(value, (list, tuple, set)):
            value = sorted(value, reverse=self.return_max_value)
            value = value[0] if len(value) > 0 else None

        return super()._deserialize(value, attr, data, **kwargs)


class DateTimeList(fields.DateTime):
    """A DateTime field that might be a list when deserializing."""

    def __init__(self, *args, **kwargs):
        """Create an instance."""
        super().__init__(*args, **kwargs)

    def _deserialize(self, value, attr, data, **kwargs):
        value = normalize_value(value)

        if isinstance(value, (list, tuple, set)):
            value = sorted(value)
            value = value[0] if len(value) > 0 else None

        return super()._deserialize(value, attr, data, **kwargs)


class Nested(fields.Nested):
    """Nested field that passes along commit info."""

    def __init__(self, *args, **kwargs):
        """Init method."""
        super().__init__(*args, **kwargs)

    @property
    def schema(self):
        """The nested ``calamus.Schema`` object.

        This method was copied from marshmallow and modified to support
        multiple different nested schemes.
        """
        if not self._schema:
            # Inherit context from parent.
            context = getattr(self.parent, "context", {})
            self._schema = {"from": {}, "to": {}}
            for nest in self.nested:
                if isinstance(nest, SchemaABC):
                    rdf_type = str(normalize_type(nest.opts.rdf_type))
                    model = nest.opts.model
                    if not rdf_type or not model:
                        raise ValueError("Both rdf_type and model need to be set on the " "schema for nested to work")
                    _schema = copy.copy(nest)
                    _schema.context.update(context)
                    # Respect only and exclude passed from parent and
                    # re-initialize fields
                    set_class = _schema.set_class
                    if self.only is not None:
                        if self._schema.only is not None:
                            original = _schema.only
                        else:  # only=None -> all fields
                            original = _schema.fields.keys()
                        _schema.only = set_class(self.only).intersection(original)
                    if self.exclude:
                        original = _schema.exclude
                        _schema.exclude = set_class(self.exclude).union(original)
                    _schema._init_fields()
                    _schema._visited = self.root._visited
                    self._schema["from"][rdf_type] = _schema
                    self._schema["to"][model] = _schema
                else:
                    if isinstance(nest, type) and issubclass(nest, SchemaABC):
                        schema_class = nest
                    elif not isinstance(nest, (str, bytes)):
                        raise ValueError("Nested fields must be passed a " "Schema, not {}.".format(nest.__class__))
                    elif nest == "self":
                        ret = self
                        while not isinstance(ret, SchemaABC):
                            ret = ret.parent
                        schema_class = ret.__class__
                    else:
                        schema_class = marshmallow.class_registry.get_class(nest)

                    rdf_type = str(normalize_type(schema_class.opts.rdf_type))
                    model = schema_class.opts.model
                    if not rdf_type or not model:
                        raise ValueError("Both rdf_type and model need to be set on the " "schema for nested to work")

                    kwargs = {}

                    self._schema["from"][rdf_type] = schema_class(
                        many=False,
                        only=self.only,
                        exclude=self.exclude,
                        context=context,
                        load_only=self._nested_normalized_option("load_only"),
                        dump_only=self._nested_normalized_option("dump_only"),
                        lazy=self.root.lazy,
                        flattened=self.root.flattened,
                        _visited=self.root._visited,
                        **kwargs,
                    )
                    self._schema["to"][model] = self._schema["from"][rdf_type]
        return self._schema



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Checks needed to determine integrity of datasets."""

import os
from collections import defaultdict

import click

from renku.command.command_builder import inject
from renku.command.util import WARNING
from renku.core import errors
from renku.core.dataset.dataset_add import add_to_dataset
from renku.core.interface.dataset_gateway import IDatasetGateway
from renku.core.migration.utils import get_pre_0_3_4_datasets_metadata
from renku.core.util import communication
from renku.core.util.os import get_safe_relative_path
from renku.domain_model.project_context import project_context


def check_dataset_old_metadata_location(**_):
    """Check location of dataset metadata.

    Args:
        _: keyword arguments.

    Returns:
        Tuple of whether dataset metadata location is valid, if an automated fix is available and string of
            found problems.
    """
    old_metadata = get_pre_0_3_4_datasets_metadata()

    if not old_metadata:
        return True, False, None

    problems = (
        WARNING + "There are metadata files in the old location."
        '\n  (use "renku migrate" to move them)\n\n\t'
        + "\n\t".join(click.style(str(path.relative_to(project_context.path)), fg="yellow") for path in old_metadata)
        + "\n"
    )

    return False, False, problems


@inject.autoparams("dataset_gateway")
def check_missing_files(dataset_gateway: IDatasetGateway, **_):
    """Find missing files listed in datasets.

    Args:
        dataset_gateway(IDatasetGateway): the dataset gateway.
        _: keyword arguments.

    Returns:
        Tuple of whether all dataset files are there, if an automated fix is available and string of found problems.
    """
    missing = defaultdict(list)

    for dataset in dataset_gateway.get_all_active_datasets():
        # NOTE: Datasets with storage backend don't have local copies of files
        if dataset.storage:
            continue
        for file_ in dataset.files:
            path = project_context.path / file_.entity.path
            file_exists = path.exists() or (file_.is_external and os.path.lexists(path))
            if not file_exists:
                missing[dataset.slug].append(file_.entity.path)

    if not missing:
        return True, False, None

    problems = WARNING + "There are missing files in datasets."

    for dataset_slug, files in missing.items():
        problems += (
            "\n\t"
            + click.style(dataset_slug, fg="yellow")
            + ":\n\t  "
            + "\n\t  ".join(click.style(path, fg="red") for path in files)
        )

    return False, False, problems


@inject.autoparams("dataset_gateway")
def check_invalid_datasets_derivation(fix, dataset_gateway: IDatasetGateway, **_):
    """Remove ``derived_from`` from import datasets.

    Args:
        fix: Whether to fix found issues.
        dataset_gateway(IDatasetGateway): the dataset gateway.
        _: keyword arguments.

    Returns:
        Tuple of whether dataset derivations are valid, if an automated fix is available and string of found problems.
    """
    invalid_datasets = []

    def fix_or_report(dataset):
        if fix:
            dataset.unfreeze()
            dataset.derived_from = None
            dataset.freeze()
            communication.info(f"Fixing dataset '{dataset.slug}'")
        else:
            invalid_datasets.append(dataset.slug)

    for dataset in dataset_gateway.get_provenance_tails():
        while dataset.derived_from is not None and dataset.derived_from.url_id is not None:
            if dataset.same_as or dataset.derived_from.url_id == dataset.id:
                fix_or_report(dataset)
                break

            try:
                dataset = dataset_gateway.get_by_id(dataset.derived_from.url_id)
            except errors.ObjectNotFoundError:
                fix_or_report(dataset)
                break

    if not invalid_datasets:
        return True, False, None

    problems = (
        WARNING
        + "There are invalid dataset metadata in the project (use 'renku doctor --fix' to fix them):"
        + "\n\n\t"
        + "\n\t".join(click.style(slug, fg="yellow") for slug in invalid_datasets)
        + "\n"
    )

    return False, True, problems


@inject.autoparams("dataset_gateway")
def check_dataset_files_outside_datadir(fix, dataset_gateway: IDatasetGateway, **_):
    """Check for dataset files that are not inside a dataset's datadir.

    Args:
        fix: Whether to fix found issues.
        dataset_gateway(IDatasetGateway): the dataset gateway.
        _: keyword arguments.

    Returns:
        Tuple of whether there are no dataset files outside of its datadir and string of found problems.
    """
    invalid_files = []
    for dataset in dataset_gateway.get_provenance_tails():
        if dataset.date_removed:
            continue

        data_dir = dataset.get_datadir()

        detected_files = []

        for file in dataset.files:
            if file.is_external or file.linked:
                continue
            try:
                get_safe_relative_path(project_context.path / file.entity.path, project_context.path / data_dir)
            except ValueError:
                detected_files.append(file)

        if not detected_files:
            continue

        if fix:
            communication.info(f"Fixing dataset '{dataset.slug}' files.")
            dataset.unfreeze()
            for file in detected_files:
                dataset.unlink_file(file.entity.path)
            dataset.freeze()
            add_to_dataset(dataset.slug, urls=[file.entity.path for file in detected_files], link=True)
        else:
            invalid_files.extend(detected_files)

    if invalid_files:
        problems = (
            WARNING
            + "There are dataset files that aren't inside their dataset's data directory "
            + "(use 'renku doctor --fix' to fix them):\n\n\t"
            + "\n\t".join(click.style(file.entity.path, fg="yellow") for file in invalid_files)
            + "\n"
        )
        return False, True, problems

    return True, False, None


@inject.autoparams("dataset_gateway")
def check_external_files(fix, dataset_gateway: IDatasetGateway, **_):
    """Find external files.

    Args:
        fix: Whether to fix found issues.
        dataset_gateway(IDatasetGateway): The injected dataset gateway.
        _: keyword arguments.

    Returns:
        Tuple of whether no external files are found, if an automated fix is available and string of found problems.
    """
    from renku.core.dataset.dataset import file_unlink

    external_files = []
    datasets = defaultdict(list)

    for dataset in dataset_gateway.get_all_active_datasets():
        for file in dataset.files:
            if file.is_external:
                external_files.append(file.entity.path)
                datasets[dataset.slug].append(file)

    if not external_files:
        return True, False, None

    external_files_str = "\n\t".join(sorted(external_files))

    if not fix:
        problems = (
            f"\n{WARNING}: External files are deprecated in favor of an external dataset backend.\n"
            "Use 'renku dataset rm' or rerun 'renku doctor' with '--fix' flag to remove them:\n\t"
            f"{external_files_str}\n"
        )
        return False, True, problems

    communication.info(
        "The following external files were deleted from the project. You need to add them later manually using a "
        f"dataset with an external storage backend:\n\t{external_files_str}"
    )

    for slug, files in datasets.items():
        file_unlink(slug=slug, yes=True, dataset_files=files)

    return True, False, None



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Base class for online repository data providers."""

import abc
import re
from pathlib import Path
from typing import TYPE_CHECKING, List

from renku.core.dataset.providers.api import ImporterApi

if TYPE_CHECKING:
    from renku.core.dataset.providers.models import DatasetAddMetadata
    from renku.domain_model.dataset import Dataset


class RepositoryImporter(ImporterApi, abc.ABC):
    """Online repository importer."""

    def download_files(self, destination: Path, extract: bool) -> List["DatasetAddMetadata"]:
        """Download dataset files from the remote provider."""
        from renku.core.dataset.providers.web import download_files

        urls, names = zip(*[(f.source, f.filename) for f in self.provider_dataset_files])

        return download_files(urls=urls, destination=destination, names=names, extract=extract)

    def tag_dataset(self, name: str) -> None:
        """Create a tag for the dataset ``name`` if the remote dataset has a tag/version."""
        from renku.core.dataset.tag import add_dataset_tag

        if self.provider_dataset.version:
            add_dataset_tag(
                dataset_slug=name,
                tag=re.sub("[^a-zA-Z0-9.-_]", "_", self.provider_dataset.version),
                description=f"Tag {self.provider_dataset.version} created by renku import",
            )

    def copy_extra_metadata(self, new_dataset: "Dataset") -> None:
        """Copy provider specific metadata once the dataset is created."""
        pass


def make_request(url, accept: str = "application/json"):
    """Execute network request."""
    from renku.core.util import requests

    response = requests.get(url, headers={"Accept": accept})
    if response.status_code != 200:
        raise LookupError(f"record not found. Status: {response.status_code}")

    return response



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Represent provenance agents."""

import re
import uuid
from typing import Optional, cast
from urllib.parse import quote

from renku.infrastructure.immutable import Slots
from renku.version import __version__, version_url


class Agent(Slots):
    """Represent executed software."""

    __slots__ = ("id", "name")

    id: str
    name: str

    def __init__(self, *, id: str, name: str, **kwargs):
        super().__init__(id=id, name=name, **kwargs)

    def __eq__(self, other):
        if self is other:
            return True
        if not isinstance(other, SoftwareAgent):
            return False
        return self.id == other.id and self.name == other.name

    def __hash__(self):
        return hash((self.id, self.name))

    @property
    def full_identity(self):
        """Return the identity of this Agent."""
        return f"{self.name} <{self.id}>"


class SoftwareAgent(Agent):
    """Represent executed software."""


# set up the default agent
RENKU_AGENT = SoftwareAgent(id=version_url, name=f"renku {__version__}")


class Person(Agent):
    """Represent a person."""

    __slots__ = ("affiliation", "alternate_name", "email")

    affiliation: Optional[str]
    alternate_name: str
    email: str

    def __init__(
        self,
        *,
        affiliation: Optional[str] = None,
        alternate_name: Optional[str] = None,
        email: Optional[str] = None,
        id: Optional[str] = None,
        name: str,
    ):
        self._validate_email(email)

        if id is None or id == "mailto:None" or id.startswith("_:"):
            full_identity = Person.get_full_identity(email, affiliation, name)
            id = Person.generate_id(email, full_identity)

        affiliation = affiliation or None
        alternate_name = alternate_name or None

        super().__init__(
            affiliation=affiliation, alternate_name=alternate_name, email=email, id=cast(str, id), name=name
        )

    def __eq__(self, other):
        if self is other:
            return True
        if not isinstance(other, Person):
            return False
        return self.id == other.id and self.full_identity == other.full_identity

    def __hash__(self):
        return hash((self.id, self.full_identity))

    @classmethod
    def from_string(cls, string):
        """Create an instance from a 'Name <email>' string."""
        regex_pattern = r"([^<>\[\]]*)" r"(?:<{1}\s*(\S+@\S+\.\S+){0,1}\s*>{1}){0,1}\s*" r"(?:\[{1}(.*)\]{1}){0,1}"
        match = re.search(regex_pattern, string)
        if match is None:
            raise ValueError(f"Couldn't parse person from string: {string}")
        name, email, affiliation = match.groups()
        if name:
            name = name.strip()
        if affiliation:
            affiliation = affiliation.strip()
        affiliation = affiliation or None

        return cls(affiliation=affiliation, email=email, name=name)

    @classmethod
    def from_dict(cls, data):
        """Create and instance from a dictionary."""
        return cls(**data)

    @staticmethod
    def generate_id(email, full_identity):
        """Generate identifier for Person."""
        # TODO: Do not use email as id
        if email:
            return f"mailto:{email}"

        id = full_identity or str(uuid.uuid4().hex)
        id = quote(id, safe="")

        return f"/persons/{id}"

    @staticmethod
    def _validate_email(email):
        """Check that the email is valid."""
        if not email:
            return
        if not isinstance(email, str) or not re.match(r"[^@]+@[^@]+\.[^@]+", email):
            raise ValueError("Email address is invalid.")

    @staticmethod
    def get_full_identity(email, affiliation, name):
        """Return name, email, and affiliation."""
        email = f" <{email}>" if email else ""
        affiliation = f" [{affiliation}]" if affiliation else ""
        return f"{name}{email}{affiliation}"

    @property
    def short_name(self):
        """Gives full name in short form."""
        names = self.name.split()
        if len(names) == 1:
            return self.name

        last_name = names[-1]
        initials = [name[0] for name in names]
        initials.pop()

        return "{}.{}".format(".".join(initials), last_name)

    @property
    def full_identity(self):
        """Return name, email, and affiliation."""
        return self.get_full_identity(self.email, self.affiliation, self.name)



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Build an execution graph for a workflow."""

from collections import defaultdict
from itertools import product
from typing import Dict, List, Tuple, Union

import networkx as nx
from networkx.algorithms.cycles import simple_cycles

from renku.core.errors import ParameterError
from renku.core.util.os import are_paths_related
from renku.domain_model.workflow import composite_plan, parameter, plan


class ExecutionGraph:
    """Represents an execution graph for one or more workflow steps."""

    def __init__(self, workflows: List["plan.AbstractPlan"], virtual_links: bool = False):
        self.workflows: List["plan.AbstractPlan"] = workflows
        self.virtual_links: List[Tuple["parameter.CommandOutput", "parameter.CommandInput"]] = []

        self.calculate_concrete_execution_graph(virtual_links=virtual_links)

    @staticmethod
    def are_paths_linked(*, output, input) -> bool:
        """Return True if input has a relation to output (i.e. can be generated by it)."""
        return output == input or are_paths_related(a=input, b=output)

    def calculate_concrete_execution_graph(self, virtual_links: bool = False):
        """Create an execution DAG between Plans showing dependencies between them.

        Resolve ParameterLink's involving ParameterMapping's to the underlying actual
        parameters and potentially also virtual links determined by parameter values.
        """
        self.graph = nx.DiGraph()
        self.virtual_links = []

        workflow_stack = self.workflows.copy()

        inputs = defaultdict(list)
        outputs = defaultdict(list)

        while workflow_stack:
            workflow = workflow_stack.pop()

            if isinstance(workflow, composite_plan.CompositePlan):
                workflow_stack.extend(workflow.plans)

                self._add_composite_plan_links_to_graph(workflow)
            else:
                if not virtual_links:
                    continue

                for input in workflow.inputs:
                    inputs[input.actual_value].append(input)
                    if not self.graph.has_edge(input, workflow):
                        self.graph.add_edge(input, workflow)

                for output in workflow.outputs:
                    outputs[output.actual_value].append(output)
                    if not self.graph.has_edge(workflow, output):
                        self.graph.add_edge(workflow, output)

        self._create_virtual_links(outputs, inputs)

    def _create_virtual_links(
        self, outputs: Dict[str, List["parameter.CommandOutput"]], inputs: Dict[str, List["parameter.CommandInput"]]
    ) -> None:
        """Add virtual links to graph based on matching inputs/outputs."""
        for output, nodes in outputs.items():
            children = []
            for input in inputs:
                if self.are_paths_linked(output=output, input=input):
                    children.extend(inputs[input])

            if not children:
                continue

            edges = product(nodes, children)

            for edge in edges:
                if not self.graph.has_edge(*edge):
                    self._add_leaf_parameter_link(*edge)
                    self.virtual_links.append(edge)

    def _add_composite_plan_links_to_graph(self, workflow: "composite_plan.CompositePlan") -> None:
        """Adds links for a grouped run to the graph."""
        if not workflow.links:
            return

        for link in workflow.links:
            for sink in link.sinks:
                self._add_leaf_parameter_link(link.source, sink)

    def _add_leaf_parameter_link(
        self, source: "parameter.CommandParameterBase", sink: "parameter.CommandParameterBase"
    ):
        """Add links between leaf parameters (resolving Mappings)."""
        if isinstance(source, parameter.ParameterMapping):
            sources = list(source.leaf_parameters)
        else:
            sources = [source]

        if isinstance(sink, parameter.ParameterMapping):
            sinks = list(sink.leaf_parameters)
        else:
            sinks = [sink]

        for param in sources + sinks:
            for workflow in self.workflows:
                wf = workflow.find_parameter_workflow(param)
                if wf:
                    break
            else:
                raise ParameterError(f"'{param.name}' is not part of any workflows.")

            edge: Union[
                Tuple["parameter.CommandParameterBase", "plan.Plan"],
                Tuple["plan.Plan", "parameter.CommandParameterBase"],
            ]

            if isinstance(param, parameter.CommandOutput):
                edge = (wf, param)
            else:
                edge = (param, wf)

            self.graph.add_edge(*edge)

        edge_list = product(sources, sinks)
        self.graph.add_edges_from(edge_list)

    @property
    def cycles(self):
        """Get potential cycles in execution graph."""
        return list(simple_cycles(self.graph))

    @property
    def workflow_graph(self) -> nx.DiGraph:
        """Return a subgraph with only workflows and their dependencies."""
        workflow_graph = nx.DiGraph()

        for node in self.graph.nodes:
            if not isinstance(node, parameter.CommandInput):
                if isinstance(node, (composite_plan.CompositePlan, plan.Plan)):
                    workflow_graph.add_node(node)
                continue

            intermediate_predecessor = next(self.graph.predecessors(node), None)
            if intermediate_predecessor is None:
                continue

            target = next(self.graph.successors(node), None)
            source = next(self.graph.predecessors(intermediate_predecessor), None)
            workflow_graph.add_edge(source, target)

        return workflow_graph



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Agents JSON-LD schemes."""

from calamus.schema import JsonLDSchema
from marshmallow import EXCLUDE, pre_load

from renku.command.schema.calamus import StringList, fields, prov, schema
from renku.domain_model.provenance.agent import Person, SoftwareAgent


class PersonSchema(JsonLDSchema):
    """Person schema."""

    class Meta:
        """Meta class."""

        rdf_type = [prov.Person, schema.Person]
        model = Person
        unknown = EXCLUDE

    @pre_load
    def fix_affiliation(self, data, **kwargs):
        """Fix affiliation to be a string."""
        affiliations = []
        affiliation = data.get("http://schema.org/affiliation")
        if affiliation:
            if not isinstance(affiliation, list):
                affiliation = [affiliation]
            for a in affiliation:
                if isinstance(a, dict):
                    name = a.get("http://schema.org/name", "")
                    if isinstance(name, list):
                        name = name[0]
                else:
                    name = str(a)
                affiliations.append(name)

            data["http://schema.org/affiliation"] = affiliations

        return data

    affiliation = StringList(schema.affiliation, load_default=None)
    alternate_name = StringList(schema.alternateName, load_default=None)
    email = fields.String(schema.email, load_default=None)
    id = fields.Id()
    name = StringList(schema.name, load_default=None)


class SoftwareAgentSchema(JsonLDSchema):
    """SoftwareAgent schema."""

    class Meta:
        """Meta class."""

        rdf_type = [prov.SoftwareAgent]
        model = SoftwareAgent
        unknown = EXCLUDE

    id = fields.Id()
    name = StringList(schema.name, load_default=None)



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Delayed controller jobs."""
import importlib

from renku.ui.service.logger import worker_log
from renku.ui.service.views.decorators import requires_cache


def to_ctrl(cache, context, user_data, renku_module, renku_ctrl):
    """Based on renku operation return valid controller."""
    ctrl_module = importlib.import_module(renku_module)
    cls = getattr(ctrl_module, renku_ctrl)

    context.pop("is_delayed", None)  # NOTE: Prevent recursive invocation.
    migrate_project = context.pop("migrate_project", False)
    return cls(cache, user_data, context, migrate_project=migrate_project)


@requires_cache
def delayed_ctrl_job(cache, context, user_data, job_id, renku_module, renku_ctrl):
    """Delayed controller job."""
    ctrl = to_ctrl(cache, context, user_data, renku_module, renku_ctrl)
    result = ctrl.to_response().json

    user = cache.ensure_user(user_data)
    worker_log.debug(f"executing delayed job {renku_ctrl} for {user.user_id}:{user.fullname}")

    user_job = cache.get_job(user, job_id)
    user_job.ctrl_result = result
    user_job.save()

    return user_job



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku service datasets unlink controller."""
from renku.command.dataset import file_unlink_command
from renku.ui.service.cache.models.job import Job
from renku.ui.service.config import MESSAGE_PREFIX
from renku.ui.service.controllers.api.abstract import ServiceCtrl
from renku.ui.service.controllers.api.mixins import RenkuOpSyncMixin
from renku.ui.service.serializers.datasets import DatasetUnlinkRequest, DatasetUnlinkResponseRPC
from renku.ui.service.views import result_response


class DatasetsUnlinkCtrl(ServiceCtrl, RenkuOpSyncMixin):
    """Controller for datasets unlink endpoint."""

    REQUEST_SERIALIZER = DatasetUnlinkRequest()
    RESPONSE_SERIALIZER = DatasetUnlinkResponseRPC()

    def __init__(self, cache, user_data, request_data, migrate_project=False):
        """Construct a datasets unlink list controller."""
        self.ctx = self.REQUEST_SERIALIZER.load(request_data)

        self.include = self.ctx.get("include_filter")
        self.exclude = self.ctx.get("exclude_filter")

        if self.include and self.exclude:
            filters = f"-I {self.include} -X {self.exclude}"
        elif not self.include and self.exclude:
            filters = f"-X {self.exclude}"
        else:
            filters = f"-I {self.include}"
        self.ctx["commit_message"] = f"{MESSAGE_PREFIX} unlink dataset {self.ctx['slug']} {filters}"

        super().__init__(cache, user_data, request_data, migrate_project=migrate_project)

    @property
    def context(self):
        """Controller operation context."""
        return self.ctx

    def renku_op(self):
        """Renku operation for the controller."""
        result = (
            file_unlink_command()
            .with_commit_message(self.ctx["commit_message"])
            .build()
            .execute(
                slug=self.ctx["slug"],
                include=self.ctx.get("include_filters"),
                exclude=self.ctx.get("exclude_filters"),
                yes=True,
            )
        )

        return result.output

    def to_response(self):
        """Execute controller flow and serialize to service response."""
        op_result, remote_branch = self.execute_and_sync()

        if isinstance(op_result, Job):
            return result_response(DatasetsUnlinkCtrl.JOB_RESPONSE_SERIALIZER, op_result)

        response = {
            "unlinked": [record.entity.path for record in op_result],
            "remote_branch": remote_branch,
            "git_url": self.ctx["git_url"],
        }

        return result_response(self.RESPONSE_SERIALIZER, response)



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Check KG structure using SHACL."""

import pyld

from renku.command.command_builder import inject
from renku.command.schema.dataset import dump_dataset_as_jsonld
from renku.command.schema.project import ProjectSchema
from renku.command.util import WARNING
from renku.core.interface.dataset_gateway import IDatasetGateway
from renku.core.util.shacl import validate_graph
from renku.domain_model.project_context import project_context


def _shacl_graph_to_string(graph):
    """Converts a shacl validation graph into human readable format.

    Args:
        graph:  SHACL validation output graph.

    Returns:
        Text describing problems found in validation.
    """
    from rdflib.namespace import Namespace
    from rdflib.term import BNode

    sh = Namespace("http://www.w3.org/ns/shacl#")

    problems = []

    for _, result in graph.subject_objects(sh.result):
        path = graph.value(result, sh.resultPath)
        res = graph.value(result, sh.resultMessage)

        if res:
            message = f"{path}: {res}"
        else:
            kind = graph.value(result, sh.sourceConstraintComponent)
            focus_node = graph.value(result, sh.focusNode)

            if isinstance(focus_node, BNode):
                focus_node = "<Anonymous>"

            message = f"{path}: Type: {kind}, Node ID: {focus_node}"

        problems.append(message)

    return "\n\t".join(problems)


def check_project_structure(**_):
    """Validate project metadata against SHACL.

    Args:
        _: keyword arguments.

    Returns:
        Tuple of whether project structure is valid, if an automated fix is available and string of found problems.
    """
    data = ProjectSchema().dump(project_context.project)

    conform, graph, t = _check_shacl_structure(data)

    if conform:
        return True, False, None

    problems = f"{WARNING}Invalid structure of project metadata\n\t{_shacl_graph_to_string(graph)}"

    return False, False, problems


@inject.autoparams("dataset_gateway")
def check_datasets_structure(dataset_gateway: IDatasetGateway, **_):
    """Validate dataset metadata against SHACL.

    Args:
        dataset_gateway(IDatasetGateway): The injected dataset gateway.
        _: keyword arguments.

    Returns:
        Tuple[bool, str]: Tuple of whether structure is valid, if an automated fix is available and of problems
            that might have been found.
    """
    ok = True

    problems = [f"{WARNING}Invalid structure of dataset metadata"]

    for dataset in dataset_gateway.get_all_active_datasets():
        data = dump_dataset_as_jsonld(dataset)
        try:
            conform, graph, t = _check_shacl_structure(data)
        except (Exception, BaseException) as e:
            problems.append(f"Couldn't validate dataset '{dataset.slug}': {e}\n\n")
            continue

        if conform:
            continue

        ok = False

        problems.append(f"{dataset.slug}\n\t{_shacl_graph_to_string(graph)}\n")

    if ok:
        return True, False, None

    return False, False, "\n".join(problems)


def _check_shacl_structure(data):
    """Validates all metadata against the SHACL schema.

    Args:
        data: JSON-LD data to validate.

    Returns:
        Validation result.
    """
    rdf = pyld.jsonld.to_rdf(data, options={"format": "application/n-quads", "produceGeneralizedRdf": True})

    return validate_graph(rdf)



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Agent view model."""


from typing import Optional

from renku.domain_model.provenance.agent import Person


class PersonViewModel:
    """View model for ``Person``."""

    def __init__(self, name: str, email: str, affiliation: Optional[str]) -> None:
        self.name = name
        self.email = email
        self.affiliation = affiliation

    @classmethod
    def from_person(cls, person: Person):
        """Create view model from ``Person``.

        Args:
            person(Person): The person to convert.
        Returns:
            View model for person
        """
        return cls(name=person.name, email=person.email, affiliation=person.affiliation)

    def __str__(self) -> str:
        email = affiliation = ""

        if self.email:
            email = f" <{self.email}>"

        if self.affiliation:
            affiliation = f" [{self.affiliation}]"

        return f"{self.name}{email}{affiliation}"



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku service worker."""
import os
from contextlib import contextmanager

import sentry_sdk
from rq import Worker
from sentry_sdk.integrations.rq import RqIntegration

from renku.core.errors import ConfigurationError, UsageError
from renku.ui.service.cache.config import REDIS_NAMESPACE
from renku.ui.service.config import SENTRY_ENABLED, SENTRY_SAMPLERATE
from renku.ui.service.jobs.queues import QUEUES, WorkerQueues
from renku.ui.service.logger import DEPLOYMENT_LOG_LEVEL, worker_log

if SENTRY_ENABLED:
    sentry_sdk.init(
        dsn=os.getenv("SENTRY_DSN"),
        environment=os.getenv("SENTRY_ENV"),
        traces_sample_rate=float(SENTRY_SAMPLERATE),
        integrations=[RqIntegration()],
    )


@contextmanager
def worker(queue_list):
    """Creates worker object."""

    def build_worker():
        """Build worker."""
        # NOTE: logging configuration has been moved to `.work(logging_level=)`
        worker_log.info(f"worker log level set to {DEPLOYMENT_LOG_LEVEL}")

        rq_worker = Worker(queue_list, connection=WorkerQueues.connection, log_job_description=False)
        worker_log.info("worker created")

        return rq_worker

    yield build_worker()


def check_queues(queue_list):
    """Check if listening on specified queues is possible."""
    for queue in queue_list:
        if queue and queue not in QUEUES:
            err_msg = "Invalid queue name: {}\n\nValid queue names: \n{}".format(queue, "\n".join(QUEUES))
            raise UsageError(err_msg)


def start_worker(queue_list):
    """Start worker."""
    q = [f"{REDIS_NAMESPACE}.{q.strip()}" for q in queue_list if q.strip()]
    check_queues(q)

    worker_log.info(f"working on queues: {q}")

    with worker(q) as rq_worker:
        worker_log.info("running worker")
        rq_worker.work(logging_level=DEPLOYMENT_LOG_LEVEL)


if __name__ == "__main__":
    queues = os.getenv("RENKU_SVC_WORKER_QUEUES")
    worker_log.info(f"working on queues: {queues}")

    if not queues:
        raise ConfigurationError(
            "Worker queues not specified. Please, set RENKU_SVC_WORKER_QUEUES environment variable."
        )

    start_worker([f"{REDIS_NAMESPACE}.{queue_name.strip()}" for queue_name in queues.strip().split(",")])



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tag management for dataset."""

import re
from typing import List, Optional, cast

from renku.command.format.dataset_tags import DATASET_TAGS_FORMATS
from renku.core import errors
from renku.core.dataset.datasets_provenance import DatasetsProvenance
from renku.core.util import communication
from renku.domain_model.dataset import Dataset, DatasetTag, Url
from renku.infrastructure.gateway.dataset_gateway import DatasetGateway
from renku.infrastructure.immutable import DynamicProxy


def add_dataset_tag(dataset_slug: str, tag: str, description="", force=False):
    """Adds a new tag to a dataset.

    Validates if the tag already exists and that the tag follows the same rules as docker tags.
    See https://docs.docker.com/engine/reference/commandline/tag/ for a documentation of docker tag syntax.

    Raises:
        ``errors.ParameterError``: If tag is too long or contains invalid characters.
    """
    if len(tag) > 128:
        raise errors.ParameterError("Tags can be at most 128 characters long.")

    if not re.match("^(?![.-])[a-zA-Z0-9_.-]{1,128}$", tag):
        raise errors.ParameterError(
            f"Tag '{tag}' is invalid.\n"
            "Only characters a-z, A-Z, 0-9, ., - and _ are allowed.\nTag can't start with a . or -"
        )
    datasets_provenance = DatasetsProvenance()
    dataset = datasets_provenance.get_by_slug(dataset_slug, strict=True)
    assert dataset is not None

    tags = datasets_provenance.get_all_tags(dataset)
    existing_tag = next((t for t in tags if t.name == tag), None)
    if existing_tag:
        if not force:
            raise errors.ParameterError(f"Tag '{tag}' already exists")
        datasets_provenance.remove_tag(dataset, existing_tag)

    new_tag = DatasetTag(dataset_id=Url(url_id=dataset.id), description=description, name=tag)

    datasets_provenance.add_tag(dataset, new_tag)


def list_dataset_tags(dataset_slug, format):
    """List all tags for a dataset."""
    datasets_provenance = DatasetsProvenance()
    dataset = datasets_provenance.get_by_slug(dataset_slug, strict=True)
    assert dataset is not None

    tags = datasets_provenance.get_all_tags(dataset)
    tags = sorted(tags, key=lambda t: t.date_created)
    tags = [cast(Dataset, DynamicProxy(t)) for t in tags]
    for tag in tags:
        tag.dataset = dataset.name

    return DATASET_TAGS_FORMATS[format](tags)


def remove_dataset_tags(dataset_slug: str, tags: List[str]):
    """Removes tags from a dataset."""
    datasets_provenance = DatasetsProvenance()
    dataset = datasets_provenance.get_by_slug(dataset_slug, strict=True)
    assert dataset is not None

    dataset_tags = datasets_provenance.get_all_tags(dataset)
    tag_names = {t.name for t in dataset_tags}
    not_found = set(tags).difference(tag_names)

    if len(not_found) > 0:
        raise errors.ParameterError(f"Tags [{', '.join(not_found)}] not found")

    for tag in dataset_tags:
        if tag.name in tags:
            datasets_provenance.remove_tag(dataset, tag)


def get_dataset_by_tag(dataset: Dataset, tag: str) -> Optional[Dataset]:
    """Return a version of dataset that has a specific tag.

    Args:
        dataset(Dataset): A dataset to return its tagged version.
        tag(str): Tag name to search for.

    Returns:
        Optional[Dataset]: The dataset pointed to by the tag or None if nothing found.
    """
    dataset_gateway = DatasetGateway()

    tags = dataset_gateway.get_all_tags(dataset)
    selected_tag = next((t for t in tags if t.name == tag), None)
    if selected_tag is None:
        return None
    return dataset_gateway.get_by_id(selected_tag.dataset_id.value)


def prompt_access_token(exporter):
    """Prompt user for an access token for a provider.

    Returns:
        The new access token
    """
    text_prompt = f"You must configure an access token\nCreate one at: {exporter.get_access_token_url()}\nAccess token"
    return communication.prompt(text_prompt, type=str)


def prompt_tag_selection(tags) -> Optional[DatasetTag]:
    """Prompt user to chose a tag or <HEAD>."""
    tags = sorted(tags, key=lambda t: t.date_created)

    text_prompt = "Tag to export: \n\n<HEAD>\t[1]\n"
    text_prompt += "\n".join(f"{t.name}\t[{i}]" for i, t in enumerate(tags, start=2))
    text_prompt += "\n\nTag"
    while True:
        selection = communication.prompt(text_prompt, type=int, default=1)

        if selection == 1:
            return None
        elif 1 < selection <= len(tags) + 1:
            return tags[selection - 2]
        communication.warn(f"{selection} is not a valid choice. Selected value has to be between 1 and {len(tags) + 1}")



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Activity JSON-LD schema."""

from marshmallow import EXCLUDE, pre_dump

from renku.command.schema.agent import PersonSchema, SoftwareAgentSchema
from renku.command.schema.annotation import AnnotationSchema
from renku.command.schema.calamus import JsonLDSchema, Nested, fields, oa, prov, renku, schema
from renku.command.schema.entity import CollectionSchema, EntitySchema
from renku.command.schema.plan import PlanSchema
from renku.command.schema.workflow_file import WorkflowFileCompositePlanSchema, WorkflowFilePlanSchema
from renku.domain_model.provenance.activity import (
    Activity,
    Association,
    Generation,
    Usage,
    WorkflowFileActivityCollection,
)
from renku.domain_model.provenance.parameter import ParameterValue


class AssociationSchema(JsonLDSchema):
    """Association schema."""

    class Meta:
        """Meta class."""

        rdf_type = prov.Association
        model = Association
        unknown = EXCLUDE

    agent = Nested(prov.agent, [SoftwareAgentSchema, PersonSchema])
    id = fields.Id()
    plan = Nested(prov.hadPlan, [PlanSchema, WorkflowFilePlanSchema, WorkflowFileCompositePlanSchema])


class UsageSchema(JsonLDSchema):
    """Usage schema."""

    class Meta:
        """Meta class."""

        rdf_type = prov.Usage
        model = Usage
        unknown = EXCLUDE

    id = fields.Id()
    # TODO: DatasetSchema, DatasetFileSchema
    entity = Nested(prov.entity, [EntitySchema, CollectionSchema])


class GenerationSchema(JsonLDSchema):
    """Generation schema."""

    class Meta:
        """Meta class."""

        rdf_type = prov.Generation
        model = Generation
        unknown = EXCLUDE

    id = fields.Id()
    # TODO: DatasetSchema, DatasetFileSchema
    entity = Nested(prov.qualifiedGeneration, [EntitySchema, CollectionSchema], reverse=True)


class ParameterValueSchema(JsonLDSchema):
    """ParameterValue schema."""

    class Meta:
        """Meta class."""

        rdf_type = [renku.ParameterValue, schema.PropertyValue]
        model = ParameterValue
        unknown = EXCLUDE

    id = fields.Id()

    parameter = fields.IRI(schema.valueReference, attribute="parameter_id")
    value = fields.Raw(schema.value)


class ActivitySchema(JsonLDSchema):
    """Activity schema."""

    class Meta:
        """Meta class."""

        rdf_type = prov.Activity
        model = Activity
        unknown = EXCLUDE

    agents = Nested(prov.wasAssociatedWith, [PersonSchema, SoftwareAgentSchema], many=True)
    annotations = Nested(oa.hasTarget, AnnotationSchema, reverse=True, many=True)
    association = Nested(prov.qualifiedAssociation, AssociationSchema)
    ended_at_time = fields.DateTime(prov.endedAtTime, format="iso", add_value_types=True)
    generations = Nested(prov.activity, GenerationSchema, reverse=True, many=True, load_default=None)
    id = fields.Id()
    invalidations = Nested(prov.wasInvalidatedBy, EntitySchema, reverse=True, many=True, load_default=None)
    parameters = Nested(
        renku.parameter,
        ParameterValueSchema,
        many=True,
        load_default=None,
    )
    path = fields.String(prov.atLocation)
    project_id = fields.IRI(renku.hasActivity, reverse=True)
    started_at_time = fields.DateTime(prov.startedAtTime, format="iso", add_value_types=True)
    usages = Nested(prov.qualifiedUsage, UsageSchema, many=True)

    @pre_dump(pass_many=True)
    def removes_ms(self, objs, many, **kwargs):
        """Remove milliseconds from datetimes.

        Note: since DateField uses `strftime` as format, which only supports timezone info without a colon
        e.g. `+0100` instead of `+01:00`, we have to deal with milliseconds manually instead of using a format string.
        """

        def _replace_times(obj):
            obj.unfreeze()
            obj.started_at_time = obj.started_at_time.replace(microsecond=0)
            obj.ended_at_time = obj.ended_at_time.replace(microsecond=0)
            obj.freeze()

        if many:
            for obj in objs:
                _replace_times(obj)
            return objs

        _replace_times(objs)
        return objs


class WorkflowFileActivityCollectionSchema(JsonLDSchema):
    """WorkflowFileActivityCollection schema."""

    class Meta:
        """Meta class."""

        rdf_type = [renku.WorkflowFileActivityCollection, prov.Collection]
        model = WorkflowFileActivityCollection
        unknown = EXCLUDE

    activities = Nested(schema.hasPart, ActivitySchema, many=True)
    agents = Nested(prov.wasAssociatedWith, [PersonSchema, SoftwareAgentSchema], many=True)
    association = Nested(prov.qualifiedAssociation, AssociationSchema)
    ended_at_time = fields.DateTime(prov.endedAtTime, add_value_types=True)
    id = fields.Id()
    project_id = fields.IRI(renku.hasActivityCollection, reverse=True)
    started_at_time = fields.DateTime(prov.startedAtTime, add_value_types=True)



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Storage factory implementation."""

from typing import TYPE_CHECKING, Dict

from renku.core.interface.storage import IStorage, IStorageFactory

if TYPE_CHECKING:
    from renku.core.dataset.providers.api import CloudStorageProviderType, ProviderCredentials


class StorageFactory(IStorageFactory):
    """Return an external storage."""

    @staticmethod
    def get_storage(
        storage_scheme: str,
        provider: "CloudStorageProviderType",
        credentials: "ProviderCredentials",
        configuration: Dict[str, str],
    ) -> "IStorage":
        """Return a storage that handles provider.

        Args:
            storage_scheme(str): Storage name.
            provider(CloudStorageProviderType): The backend provider.
            credentials(ProviderCredentials): Credentials for the provider.
            configuration(Dict[str, str]): Storage-specific configuration that are passed to the IStorage implementation

        Returns:
            An instance of IStorage.
        """
        from .rclone import RCloneStorage

        return RCloneStorage(
            storage_scheme=storage_scheme,
            provider=provider,
            credentials=credentials,
            provider_configuration=configuration,
        )



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Commands to launch service components."""
import os
import signal
import subprocess
import sys
import tempfile
import time
from datetime import datetime
from pathlib import Path

import click
import psutil

import renku.ui.cli.utils.color as color
from renku.command.util import ERROR

RENKU_DAEMON_LOG_FILE = "renku.log"
RENKU_DAEMON_ERR_FILE = "renku.err"

SERVICE_COMPONENT_TAGS = ["api", "worker"]


def run_api(addr="0.0.0.0", port=8080, timeout=600):
    """Run service JSON-RPC API."""
    from gunicorn.app.wsgiapp import run

    svc_num_workers = os.getenv("RENKU_SVC_NUM_WORKERS", "1")
    svc_num_threads = os.getenv("RENKU_SVC_NUM_THREADS", "2")

    svc_timeout = int(os.getenv("REQUEST_TIMEOUT", timeout))

    loading_opt = "--preload"

    sys.argv = [
        "gunicorn",
        "renku.ui.service.entrypoint:app",
        loading_opt,
        "-c",
        "gunicorn.conf.py",
        "-b",
        f"{addr}:{port}",
        "--timeout",
        f"{svc_timeout}",
        "--workers",
        svc_num_workers,
        "--worker-class",
        "gthread",
        "--threads",
        svc_num_threads,
        "--log-level",
        "debug",
    ]

    sys.exit(run())


def run_worker(queues):
    """Run service workers."""
    from renku.ui.service.jobs.queues import QUEUES
    from renku.ui.service.worker import start_worker

    if not queues:
        queues = os.getenv("RENKU_SVC_WORKER_QUEUES", "")
        queues = [queue_name.strip() for queue_name in queues.strip().split(",") if queue_name.strip()]

        if not queues:
            queues = QUEUES

    start_worker(queues)


def check_cmdline(cmdline, include=None):
    """Check `cmdline` command of a process."""
    include = include or []
    service_components = include + SERVICE_COMPONENT_TAGS

    for cmd in service_components:
        if cmd in cmdline:
            return True

    return False


def is_renku_process(process, include):
    """Return true if this is a renku process."""
    process_name = process.name().lower()

    if process_name == "renku" and check_cmdline(process.cmdline(), include):
        return True
    elif "python" not in process_name:
        return False

    try:
        command_line = process.cmdline()
        if not check_cmdline(command_line, include):
            return False

        for line in command_line:
            if line.endswith("renku"):
                return True

    except (psutil.AccessDenied, psutil.NoSuchProcess):
        pass

    return False


def list_renku_processes(include=None):
    """List renku processes."""
    include = include or []

    renku_processes_all = []
    for pid in sorted(psutil.pids()):
        try:
            proc = psutil.Process(pid)

            if is_renku_process(proc, include) and proc.status() != "zombie":
                renku_processes_all.append(proc)

        except (psutil.NoSuchProcess, psutil.ZombieProcess):
            continue

    renku_proc_info = sorted(
        [
            {
                "create_time": datetime.fromtimestamp(proc.create_time()).strftime("%d.%b %H:%M"),
                "pid": proc.pid,
                "cmdline": f"renku {' '.join(proc.cmdline()[2:])}",
                "status": proc.status(),
                "mem_perct": proc.memory_percent(),
                "cpu_perct": proc.cpu_percent(),
                "num_threads": proc.num_threads(),
            }
            for proc in renku_processes_all
        ],
        key=lambda k: k["cmdline"],
    )

    return renku_proc_info


def read_logs(log_file, follow=True, output_all=False):
    """Read logs file. Supports following logs in realtime."""
    if follow and not output_all:
        log_file.seek(0, os.SEEK_END)

    while True:
        line = log_file.readline()
        if not line and follow:
            time.sleep(0.1)
            continue

        if not line and not follow:
            return

        yield line


@click.group()
@click.option("-e", "--env", default=None, type=click.Path(exists=True, dir_okay=False), help="Path to the .env file.")
@click.pass_context
def service(ctx, env):
    """Manage service components."""
    import redis
    import rq  # noqa: F401
    from dotenv import load_dotenv

    if ctx.invoked_subcommand in ["apispec", "logs", "api"]:
        return  # Redis not needed

    try:
        from renku.ui.service.cache.base import BaseCache

        BaseCache.cache.ping()

        load_dotenv(dotenv_path=env)
    except ImportError:
        # NOTE: Service dependency is missing.

        click.echo(
            ERROR + "Dependency not found! "
            "Please install `pip install renku[service]` to enable service component control."
        )

        ctx.exit(1)

    except redis.exceptions.ConnectionError as e:
        # NOTE: Cannot connect to the service dependencies, ie. Redis.
        click.echo(ERROR + f"Cannot connect to Redis: {e}")

        ctx.exit(1)


@service.command(name="api")
@click.option(
    "-a",
    "--addr",
    type=str,
    default="0.0.0.0",
    show_default=True,
    help="Address on which API service should listen to. By default uses IPv4.",
)
@click.option(
    "-p",
    "--port",
    type=int,
    default=8080,
    show_default=True,
    help="Port on which API service should listen to. Avoid ports below 1024, for those use reverse-proxies.",
)
@click.option(
    "-t",
    "--timeout",
    type=int,
    default=600,
    show_default=True,
    help="Request silent for more than this many seconds are dropped.",
)
def api_start(addr, port, timeout):
    """Start service JSON-RPC API in active shell session."""
    run_api(addr, port, timeout)


@service.command(name="worker")
@click.option("-q", "--queue", multiple=True)
def worker_start(queue):
    """Start service worker in active shell session. By default it listens on all queues."""
    run_worker([q.strip() for q in queue if q])


@service.command(name="ps")
@click.pass_context
def ps(ctx):
    """Check status of running services."""
    from renku.core.util.tabulate import tabulate

    processes = list_renku_processes()
    headers = [{k.upper(): v for k, v in rec.items()} for rec in processes]

    output = tabulate(processes, headers=headers)

    if not processes:
        click.echo("Renku service components are down.")
        ctx.exit()

    click.echo(output)


@service.command(name="up")
@click.option("-d", "--daemon", is_flag=True, default=False, help="Starts all processes in daemon mode.")
@click.option("-rd", "--runtime-dir", default=".", help="Directory for runtime metadata in daemon mode.")
@click.pass_context
def all_start(ctx, daemon, runtime_dir):
    """Start all service components."""
    from circus import get_arbiter

    from renku.core.util.contexts import chdir

    services = [
        {
            "name": "RenkuCoreService",
            "cmd": "renku",
            "args": ["service", "api"],
            "numprocesses": 1,
            "env": os.environ.copy(),
            "shell": True,
        },
        {
            "name": "RenkuCoreWorker",
            "cmd": "renku",
            "args": ["service", "worker"],
            "numprocesses": 1,
            "env": os.environ.copy(),
            "shell": True,
        },
    ]

    def launch_arbiter(arbiter):
        """Helper for launching arbiter process."""
        with chdir(runtime_dir):
            try:
                arbiter.start()
            finally:
                arbiter.stop()

    if not daemon:
        launch_arbiter(get_arbiter(services))
        ctx.exit()

    # NOTE: If we are running in daemon mode, the runtime directory is generated is OS /tmp directory.
    # Since in this case daemon is long running process we don't want to pollute user space.
    if not runtime_dir or runtime_dir == ".":
        runtime_dir = tempfile.mkdtemp()

    os.environ["CACHE_DIR"] = runtime_dir
    click.echo(f"Using runtime directory: {runtime_dir}")

    log_stdout = Path(runtime_dir) / RENKU_DAEMON_LOG_FILE
    log_stderr = Path(runtime_dir) / RENKU_DAEMON_ERR_FILE

    subprocess.Popen(
        ["renku", "service", "up", "--runtime-dir", runtime_dir],
        stdout=log_stdout.open(mode="w"),
        stderr=log_stderr.open(mode="w"),
        start_new_session=True,
    )

    click.secho("OK", fg=color.GREEN)


@service.command(name="down")
def all_stop():
    """Stop all service components."""
    # NOTE: We include `renku service up` because that process contains the arbiter and watcher.
    processes = list_renku_processes(["up"])

    for proc in processes:
        click.echo(f"Shutting down [{proc['pid']}] `{proc['cmdline']}`")
        try:
            os.kill(proc["pid"], signal.SIGKILL)
        except ProcessLookupError:
            click.echo(f"Process [{proc['pid']}] `{proc['cmdline']}` not found - skipping")
            continue

    if processes:
        click.secho("OK", fg=color.GREEN)
    else:
        click.echo("Nothing to shut down.")


@service.command(name="restart")
def all_restart():
    """Restart all running service components."""
    processes = list_renku_processes()

    for proc in processes:
        click.echo(f"Restarting `{proc['cmdline']}`")
        os.kill(proc["pid"], signal.SIGKILL)

    if processes:
        click.secho("OK", fg=color.GREEN)
    else:
        click.echo("Nothing to restart.")


@service.command(name="logs")
@click.option("-f", "--follow", is_flag=True, default=False, help="Follows logs of damonized service components.")
@click.option(
    "-a", "--output-all", is_flag=True, default=False, help="Outputs ALL logs of damonized service components."
)
@click.option("-e", "--errors", is_flag=True, default=False, help="Outputs all errors of damonized service components.")
@click.pass_context
def all_logs(ctx, follow, output_all, errors):
    """Check logs of all running daemonized service components."""
    processes = list_renku_processes(["up"])

    if not processes:
        click.echo("Daemonized component processes are not running.\nStart them with `renku service up --daemon`")
        ctx.exit()

    for proc in processes:
        if "cmdline" in proc and "up" in proc["cmdline"]:
            runtime_dir = Path(proc["cmdline"].split("--runtime-dir")[-1].strip())

            stream = runtime_dir / RENKU_DAEMON_LOG_FILE
            if errors:
                stream = runtime_dir / RENKU_DAEMON_ERR_FILE

            for line in read_logs(stream.open(mode="r"), follow=follow, output_all=output_all):
                click.echo(line)


@service.command(name="apispec")
def apispec():
    """Return the api spec."""
    from renku.ui.service.entrypoint import app
    from renku.ui.service.views.apispec import get_apispec

    with app.test_request_context():
        click.echo(get_apispec(app).to_yaml())



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku service exceptions."""
import os
from urllib.parse import urlparse

from renku.ui.service.config import (
    DOCS_URL_BASE,
    DOCS_URL_ERRORS,
    ERROR_NOT_AVAILABLE,
    SENTRY_ENABLED,
    SVC_ERROR_INTERMITTENT,
    SVC_ERROR_PROGRAMMING,
    SVC_ERROR_USER,
)
from renku.ui.service.serializers.headers import OptionalIdentityHeaders


class ServiceError(Exception):
    """Error handler for service errors.

    Set the documentation reference for developers willing to explore the base code.
    Handle Sentry, adding user data (when available) and error details.
    """

    devReference = f"{DOCS_URL_BASE}{DOCS_URL_ERRORS}" + "{class_name}"

    def __init__(
        self,
        code=None,
        userMessage=None,
        devMessage=None,
        userReference=None,
        devReference=None,
        exception=None,
    ):
        """Initialize the service error."""
        super().__init__()
        self.exception = exception
        self.devReference = f"{DOCS_URL_BASE}{DOCS_URL_ERRORS}{self.__class__.__name__}"
        if exception:
            self.__cause__ = exception

        if code:
            self.code = code
        if userMessage:
            self.userMessage = userMessage
        if devMessage:
            self.devMessage = devMessage
        if userReference:
            self.userReference = userReference
        if devReference:
            self.devReference = devReference

        if SENTRY_ENABLED:
            self._handle_sentry()

    def _get_user(self):
        """Get the user detail when the identity headers are set."""
        from flask import request

        user_identity = OptionalIdentityHeaders().load(request.headers)
        if user_identity.get("user_id", False):
            return {
                "logged": True,
                "id": user_identity["user_id"],
                "email": user_identity.get("email", "N/A"),
                "fullname": user_identity.get("fullname", "N/A"),
            }
        return {"logged": False}

    def _handle_sentry(self):
        """Send the expection to Sentry when available."""
        from sentry_sdk import capture_exception, set_context, set_tag, set_user

        if SENTRY_ENABLED:
            # NOTE: Trap the process working directory when internal error occurs.
            try:
                path = os.readlink(f"/proc/{os.getpid()}/cwd")
                set_context("pwd", {"path": path})
            except (Exception, BaseException):
                pass

            # NOTE: try to set user details when available
            user = self._get_user()
            set_user(user)

            # NOTE: set erro code tag and details section
            set_tag("error_code", self.code)
            details = {"code": self.code}
            if hasattr(self, "devMessage"):
                details["developer_message"] = self.devMessage
            if hasattr(self, "devReference"):
                details["developer_reference"] = self.devReference
            if hasattr(self, "userMessage"):
                details["user_message"] = self.userMessage
            if hasattr(self, "userReference"):
                details["user_reference"] = self.userReference
            set_context("details", details)

            # NOTE: add the sentry URL to the exception
            try:
                sentry = capture_exception(self)
                sentry_dsn = os.getenv("SENTRY_DSN", None)
                if sentry_dsn is not None:
                    try:
                        sentry_url = urlparse(sentry_dsn)
                        sentry_target = sentry_url.netloc.split("@")[-1]
                        # NOTE: sentry doesn't support a global search. A proper link would require the specific org
                        sentry = f"{sentry_url.scheme }://{sentry_target}/organizations/sentry?query={sentry}"
                    except Exception:  # nosec
                        pass
            except KeyError as e:
                sentry = f"Unexpected error while reporting to Sentry: {str(e)}"
        else:
            sentry = "Unavailable"

        self.sentry = sentry


class UserInvalidGenericFieldsError(ServiceError):
    """One or more fields provided by the user has an invalid value.

    This is a generic error. Where possible, it would be best to create specific errors meaningful for the users.
    """

    code = SVC_ERROR_USER + 1
    userMessage = "Some of the provided values are wrong: {wrong_values}"
    devMessage = "Wrong values have been provided: {wrong_values}"

    def __init__(self, exception=None, wrong_values=ERROR_NOT_AVAILABLE):
        super().__init__(
            userMessage=self.userMessage.format(wrong_values=wrong_values),
            devMessage=self.devMessage.format(wrong_values=wrong_values),
            exception=exception,
        )


class UserRepoUrlInvalidError(ServiceError):
    """The provided URL is not a valid Git repository.

    This usually happens when the URL is wrong.

    There are many possibilities:

    * The URL is wrong
    * The target is not a valid git repository
    * The user doesn't have the required permissions to access the repository
    * The target may temporarily be unavailable.
    """

    code = SVC_ERROR_USER + 10
    userMessage = (
        "A valid Git repository is not available at the target URL, "
        "or the user does not have the required permissions to access it."
    )
    devMessage = "Repository not found. Git error message: {error_message}"

    def __init__(self, exception=None, error_message=ERROR_NOT_AVAILABLE):
        super().__init__(devMessage=self.devMessage.format(error_message=error_message), exception=exception)


class UserRepoNoAccessError(ServiceError):
    """The target repository cannot be accessed by the current user.

    This is usually due to lack of permissions by the user, although in that case
    the repository manager usually returns a Not Found error instead.
    """

    code = SVC_ERROR_USER + 11
    userMessage = "Error accessing the repository due to lack of permissions."
    devMessage = "Access denied to the repository. Git error message: {error_message}"

    def __init__(self, exception=None, error_message=ERROR_NOT_AVAILABLE):
        super().__init__(devMessage=self.devMessage.format(error_message=error_message), exception=exception)


class UserRepoBranchInvalidError(ServiceError):
    """The provided URL is a valid Git repository, but the target branch does not exist.

    This usually happens when the target branch does not exist anymore.
    It is possible it was there at some point and it has been removed in the meanwhile.
    """

    code = SVC_ERROR_USER + 12
    userMessage = "The target URL is a valid Git repository, but we could not find the branch '{branch}'."
    devMessage = "Branch not valid. Git error message: {error_message}"

    def __init__(self, exception=None, branch=ERROR_NOT_AVAILABLE, error_message=ERROR_NOT_AVAILABLE):
        super().__init__(
            userMessage=self.userMessage.format(branch=branch),
            devMessage=self.devMessage.format(error_message=error_message),
            exception=exception,
        )


class UserRepoReferenceInvalidError(ServiceError):
    """The provided URL is a valid Git repository, but the target reference does not exist.

    This usually happens when the target branch, tag or commit does not exist anymore.
    It's possible it was there and the history has been re-written in the meanwhile, or it
    was never there at all.
    """

    code = SVC_ERROR_USER + 13
    userMessage = (
        "The target URL is a valid Git repository, but we could not find the reference"
        "'{reference}'. Please specify a valid tag, branch or commit"
    )
    devMessage = "Reference not valid. Git error message: {error_message}"

    def __init__(self, exception=None, branch=ERROR_NOT_AVAILABLE, error_message=ERROR_NOT_AVAILABLE):
        super().__init__(
            userMessage=self.userMessage.format(branch=branch),
            devMessage=self.devMessage.format(error_message=error_message),
            exception=exception,
        )


class UserAnonymousError(ServiceError):
    """The user must login in to use the target endpoint."""

    code = SVC_ERROR_USER + 30
    userMessage = "It's necessary to be authenticated to access the target data."
    devMessage = "User identification is incorrect or missing."

    def __init__(self, exception=None):
        super().__init__(exception=exception)


class UserMissingFieldError(ServiceError):
    """The user has not provided an expected field."""

    code = SVC_ERROR_USER + 40
    userMessage = "The following field is required: {field}."
    devMessage = "User did not provide the mandatory field '{field}'."

    def __init__(self, exception=None, field=ERROR_NOT_AVAILABLE):
        super().__init__(
            userMessage=self.userMessage.format(field=field),
            devMessage=self.devMessage.format(field=field),
            exception=exception,
        )


class UserTemplateInvalidError(ServiceError):
    """The provided URL doesn't lead to a valid template repository."""

    code = SVC_ERROR_USER + 101
    userMessage = "The target repository is not a valid Renku template repository: {error_message}"
    devMessage = "Target repository is not a valid template.: {error_message}"

    def __init__(self, exception=None, error_message=ERROR_NOT_AVAILABLE):
        super().__init__(
            userMessage=self.userMessage.format(error_message=error_message),
            devMessage=self.devMessage.format(error_message=error_message),
            exception=exception,
        )


class UserProjectCreationError(ServiceError):
    """Error when creating a new project from a Renku template.

    One (or more) project field is wrong.
    """

    code = SVC_ERROR_USER + 102
    userMessage = "There is an error with a project field: {error_message}."
    devMessage = "Project creation from a Renku template failed. The user provided wrong field(s): {error_message}."

    def __init__(self, exception=None, error_message=ERROR_NOT_AVAILABLE):
        super().__init__(
            userMessage=self.userMessage.format(error_message=error_message),
            devMessage=self.devMessage.format(error_message=error_message),
            exception=exception,
        )


class UserNonRenkuProjectError(ServiceError):
    """The target repository is valid but it is not a Renku project."""

    code = SVC_ERROR_USER + 110
    userMessage = "The target project is not a Renku project."
    devMessage = "Cannot work on a non Renku project."

    def __init__(self, exception=None):
        super().__init__(exception=exception)


class UserProjectMetadataCorruptError(ServiceError):
    """The metadata in the project os corrupt and couldn't be loaded."""

    code = SVC_ERROR_USER + 120
    userMessage = "The Renku metadata in the project is corrupt and couldn't be loaded: {error_message}"
    devMessage = "Couldn't parse project metadata due to a JSON parsing error: {error_message}"

    def __init__(self, exception=None, error_message=ERROR_NOT_AVAILABLE):
        super().__init__(
            userMessage=self.userMessage.format(error_message=error_message),
            devMessage=self.devMessage.format(error_message=error_message),
            exception=exception,
        )


class UserDatasetsMultipleImagesError(ServiceError):
    """Multiple images dataset have the same priority."""

    code = SVC_ERROR_USER + 130
    userMessage = "Multiple dataset images must have different priorities."
    devMessage = "The input parameters contain multiple dataset images with the same priority."

    def __init__(self, exception=None):
        super().__init__(exception=exception)


class UserDatasetsUnreachableImageError(ServiceError):
    """Dataset image not reachable."""

    code = SVC_ERROR_USER + 131
    userMessage = "Cannot download the dataset image."
    devMessage = "The remote dataset image is not reachable."

    def __init__(self, exception=None):
        super().__init__(exception=exception)


class UserDatasetsUnlinkError(ServiceError):
    """Dataset unlink parameters not valid."""

    code = SVC_ERROR_USER + 132
    userMessage = "Please provide valid unlink parameter."
    devMessage = "Unlink parameters not valid."

    def __init__(self, exception=None):
        super().__init__(exception=exception)


class UserDatasetsNotFoundError(ServiceError):
    """Dataset couldn't be found in project."""

    code = SVC_ERROR_USER + 133
    userMessage = (
        "The dataset doesn't exist in the project. Please check your inputs or create the target dataset first."
    )
    devMessage = "The dataset couldn't be found in the project."

    def __init__(self, exception=None):
        super().__init__(exception=exception)


class UserOutdatedProjectError(ServiceError):
    """The operation can be done only after updating the target project."""

    code = SVC_ERROR_USER + 140
    userMessage = "This operation requires to update the project first."
    devMessage = "The requested operation can only be performed on an up-to-date project."

    def __init__(self, exception=None):
        super().__init__(exception=exception)


class UserNewerRenkuProjectError(ServiceError):
    """The target repository is valid but it needs a newer renku version than is currently deployed."""

    code = SVC_ERROR_USER + 141
    userMessage = (
        "The target project requires renku version {minimum_version}, but this deployment is on "
        "version {current_version}."
    )
    devMessage = "Update the deployed core-service version to support newer projects."

    def __init__(self, exception=None, minimum_version=ERROR_NOT_AVAILABLE, current_version=ERROR_NOT_AVAILABLE):
        super().__init__(
            userMessage=self.userMessage.format(minimum_version=minimum_version, current_version=current_version),
            exception=exception,
        )


class UserProjectTemplateReferenceError(ServiceError):
    """The project's template original reference cannot be found anymore.

    The reference has probably been removed, either on purpose or as a side effect of a
    forced push.
    """

    code = SVC_ERROR_USER + 141
    userMessage = (
        "The project's template original reference has been removed or overwritten."
        " Manually changing it in a session may fix the problem."
        " Further details: {message}."
    )
    devMessage = "Template reference is not available anymore. Details: {message}."

    def __init__(self, exception):
        super().__init__(
            userMessage=self.userMessage.format(message=str(exception)),
            devMessage=self.devMessage.format(message=str(exception)),
            exception=exception,
        )


class UserUploadTooLargeError(ServiceError):
    """The user tried to upload a file that is too large.

    Maximum upload size can be set with the ``maximumUploadSizeBytes`` chart value or ``MAX_CONTENT_LENGTH``
    environment value.
    """

    code = SVC_ERROR_USER + 150
    userMessage = "The file you are trying to upload is too large. Maximum allowed size is: {maximum_size}"
    devMessage = "Uploaded file size was larger than ``MAX_CONTENT_LENGTH``."

    def __init__(self, exception, maximum_size: str):
        super().__init__(
            userMessage=self.userMessage.format(maximum_size=maximum_size),
            devMessage=self.devMessage.format(maximum_size=maximum_size),
            exception=exception,
        )


class ProgramInvalidGenericFieldsError(ServiceError):
    """One or more fields are unexpected.

    This error should not be triggered by any user input, but rather by unexpected fields.
    It is most likely a bug on the client side.
    """

    code = SVC_ERROR_PROGRAMMING + 1
    userMessage = "There was an unexpected error while handling project data."
    devMessage = "Unexpected fields have been provided: {wrong_values}"

    def __init__(self, exception=None, wrong_values=ERROR_NOT_AVAILABLE):
        super().__init__(
            devMessage=self.devMessage.format(wrong_values=wrong_values),
            exception=exception,
        )


class ProgramRepoUnknownError(ServiceError):
    """Unknown error when working with the repository.

    This is a fallback error and it should ideally never show up.
    When it happens, it's best to investigate the underlying error and either
    create a more specific error or open an issue.
    """

    code = SVC_ERROR_PROGRAMMING + 10
    userMessage = "Fatal error occurred while working on the repository."
    devMessage = "Unexpected repository error. Git error message: {error_message}"

    def __init__(self, exception=None, error_message=ERROR_NOT_AVAILABLE):
        super().__init__(devMessage=self.devMessage.format(error_message=error_message), exception=exception)


class ProgramGitError(ServiceError):
    """Unknown error when working with git.

    This is a fallback error and it should ideally never show up.
    When it happens, it's best to investigate the underlying error and either
    create a more specific error or open an issue.
    """

    code = SVC_ERROR_PROGRAMMING + 20
    userMessage = "Fatal error occurred while processing a git operation on the repository."
    devMessage = "Unexpected git error. Git error message: {error_message}"

    def __init__(self, exception=None, error_message=ERROR_NOT_AVAILABLE):
        super().__init__(devMessage=self.devMessage.format(error_message=error_message), exception=exception)


class ProgramContentTypeError(ServiceError):
    """Content type not correctly specified."""

    code = SVC_ERROR_PROGRAMMING + 30
    userMessage = "Our servers could not process the received data."
    devMessage = (
        "Invalid requests headers. The content type should be set to '{expected_type}' instead of '{content_type}'"
    )

    def __init__(self, content_type=ERROR_NOT_AVAILABLE, expected_type=ERROR_NOT_AVAILABLE):
        super().__init__(devMessage=self.devMessage.format(content_type=content_type, expected_type=expected_type))


class ProgramProjectCreationError(ServiceError):
    """Error when creating a new project due to one or more wrong fields not controlled by the user."""

    code = SVC_ERROR_PROGRAMMING + 102
    userMessage = "Our servers could not create the new project."
    devMessage = (
        "Project creation from a Renku template failed."
        "It is likely that some non user input fields contain unexpected values: {error_message}."
    )

    def __init__(self, exception=None, error_message=ERROR_NOT_AVAILABLE):
        super().__init__(
            devMessage=self.devMessage.format(error_message=error_message),
            exception=exception,
        )


class ProgramProjectCorruptError(ServiceError):
    """Error triggered by an unexpected project corruption, not easily fixable by the user."""

    code = SVC_ERROR_PROGRAMMING + 120
    userMessage = "The Renku project seems partially corrupted, and the operation could not finish."
    devMessage = "Project corrupted. It probably requires manually fixing it, or reverting to a previous commit."

    def __init__(self, exception=None):
        super().__init__(exception=exception)


class ProgramGraphCorruptError(ServiceError):
    """Error triggered by an unexpected failure while exporting the graph ."""

    code = SVC_ERROR_PROGRAMMING + 121
    userMessage = "The Renku project seems partially corrupted, and the graph could not be generated."
    devMessage = "Could not generate the project graph. Check the Sentry exception for further details."

    def __init__(self, exception=None):
        super().__init__(exception=exception)


class ProgramUpdateProjectError(ServiceError):
    """Updating the project failed."""

    code = SVC_ERROR_USER + 140
    userMessage = "Our servers could not update the project succesfully. You could try doing it manually in a session."
    devMessage = "Updating the target project failed. Check the Sentry exception for further details."
    userReference = "https://renku.readthedocs.io/en/stable/how-to-guides/general/upgrading-renku.html"

    def __init__(self, exception=None):
        super().__init__(exception=exception)


class ProgramRenkuError(ServiceError):
    """Renku error unexpected at service level for a specific operation."""

    code = SVC_ERROR_PROGRAMMING + 200
    userMessage = "Our servers could not process the requested operation. The following details may help: {renku_error}"
    devMessage = "Unexpected Renku exception from a service operation: {renku_error}"

    def __init__(self, exception=None):
        message = str(exception)
        super().__init__(
            userMessage=self.userMessage.format(renku_error=message),
            devMessage=self.devMessage.format(renku_error=message),
            exception=exception,
        )


class ProgramHttpMethodError(ServiceError):
    """HTTP error 405 method not allowed.

    The method cannot be used on the target endpoint. The service only supports GET and POST.
    """

    code = SVC_ERROR_PROGRAMMING + 405
    userMessage = "One of the resources on our servers could not process the data properly."
    devMessage = (
        "Error 405 - method not allowed. Check if you are using GET or POST as specified in the "
        "API documentation. Mind that the service doesn't support any other method."
    )

    def __init__(self, exception=None):
        super().__init__(exception=exception)


class ProgramHttpMissingError(ServiceError):
    """HTTP error 404 not found.

    Either the URL is wrong or the user doesn't have permissions to access it.
    """

    code = SVC_ERROR_PROGRAMMING + 404
    userMessage = "One of the resources on our servers is not available at the moment."
    devMessage = (
        "Error 404 - not found. The resources are currently unavailable. "
        "This may happen when the user doesn't have permissions to access the target endpoint, "
        "or the endpoint URL is wrong or has been renamed."
    )

    def __init__(self, exception=None):
        super().__init__(exception=exception)


class ProgramHttpRequestError(ServiceError):
    """HTTP error 400 bad request.

    This is usually triggered by wrong parameters or payload. Double check the API documentation.
    """

    code = SVC_ERROR_PROGRAMMING + 400
    userMessage = "One of the resources on our servers could not process the data properly."
    devMessage = "Error 400 - bad request. Check if the payload and the parameters are correctly structured."

    def __init__(self, exception=None):
        super().__init__(exception=exception)


class ProgramHttpTimeoutError(ServiceError):
    """HTTP error 408 request timeout.

    This is usually triggered by wrong parameters or payload. Double check the API documentation.
    """

    code = SVC_ERROR_PROGRAMMING + 408
    userMessage = "The request took too long and was interruped."
    devMessage = "Error 408 - request timeout. There may be a client side delay."

    def __init__(self, exception=None):
        super().__init__(exception=exception)


class ProgramHttpServerError(ServiceError):
    """Any other HTTP error.

    Most of the times (always?) this will be a 50x error.
    The only way to get more details is to access the Sentry exception.
    """

    code = SVC_ERROR_PROGRAMMING
    userMessage = "Fatal error occurred while working on the repository."
    devMessage = "Unexpected repository error. Check the Sentry exception for further details."

    def __init__(self, exception=None, http_error_code=599):
        super().__init__(exception=exception, code=self.code + http_error_code)


class ProgramInternalError(ServiceError):
    """Unknown internal error.

    This is an unexpected exception probably triggered at the core level.
    Please use Sentry to get more information.
    """

    code = SVC_ERROR_PROGRAMMING + 900
    userMessage = "Our servers generated an unexpected error while processing data."
    devMessage = "Renku service internal error. Further information: {error_message}"

    def __init__(self, exception=None, error_message=ERROR_NOT_AVAILABLE):
        super().__init__(devMessage=self.devMessage.format(error_message=error_message), exception=exception)


class IntermittentProjectIdError(ServiceError):
    """The project id cannot be found in the cache.

    This is an unexpected error possibly related to a cache malfunction, or just an unlucky case if the operation
    was attempted just after the project was cleaned up form the local cache. The latter is unlikely to happen.
    The culprit may be the client if it explicitly provides the project id.
    """

    code = SVC_ERROR_INTERMITTENT + 1
    userMessage = "An unexpected error occurred. This may be a temporary problem. Please try again in a few minutes."
    devMessage = (
        "Project id cannot be found. It may be a temporary problem. Check the Sentry exception for further details."
    )

    def __init__(self, exception=None):
        super().__init__(exception=exception)


class IntermittentAuthenticationError(ServiceError):
    """The user credentials were received but they are not valid.

    This may happen for a number of reasons. Triggering a new login will likely fix it.
    """

    code = SVC_ERROR_INTERMITTENT + 30
    userMessage = "Invalid user credentials. Please try to log out and in again."
    devMessage = "Authentication error. Check the Sentry exception to inspect the headers"

    def __init__(self, exception=None):
        super().__init__(exception=exception)


class IntermittentFileExistsError(ServiceError):
    """An operation failed because one or more files were expected not to exist.

    It may be a synchronization error happening when two or more concurrent operations overlap
    and one creates content unexpected from another one.
    """

    code = SVC_ERROR_INTERMITTENT + 110
    userMessage = "There was an error with the file '{file_name}'. Please refresh the page and try again."
    devMessage = "Unexpected error on file '{file_name}', possibly caused by concurrent actions."

    def __init__(self, exception=None, file_name=ERROR_NOT_AVAILABLE):
        super().__init__(
            userMessage=self.userMessage.format(file_name=file_name),
            devMessage=self.devMessage.format(file_name=file_name),
            exception=exception,
        )


class IntermittentFileNotExistsError(ServiceError):
    """An operation failed because one or more files were expected to exist.

    It may be a synchronization error happening when two or more concurrent operations overlap
    and one consumes the file expected by another one.
    """

    code = SVC_ERROR_INTERMITTENT + 111
    userMessage = "There was an error with the file '{file_name}'. Please refresh the page and try again."
    devMessage = "Unexpected error on file '{file_name}', possibly caused by concurrent actions."

    def __init__(self, exception=None, file_name=ERROR_NOT_AVAILABLE):
        super().__init__(
            userMessage=self.userMessage.format(file_name=file_name),
            devMessage=self.devMessage.format(file_name=file_name),
            exception=exception,
        )


class IntermittentSettingExistsError(ServiceError):
    """An operation failed because one or more project settings does not exist.

    It may be a synchronization error happening when two or more concurrent operations overlap
    and one tries to work on content already deleted from another one.
    """

    code = SVC_ERROR_INTERMITTENT + 112
    userMessage = "There was an error with the setting '{setting_name}'. Please refresh the page and try again."
    devMessage = "Unexpected error on setting '{setting_name}', possibly caused by concurrent actions."

    def __init__(self, exception=None, setting_name=ERROR_NOT_AVAILABLE):
        super().__init__(
            userMessage=self.userMessage.format(setting_name=setting_name),
            devMessage=self.devMessage.format(setting_name=setting_name),
            exception=exception,
        )


class IntermittentDatasetExistsError(ServiceError):
    """An operation failed because a dataset was expected not to exist.

    It may be a synchronization error happening when two or more concurrent operations overlap
    and one tries to create content already created from another one.
    """

    code = SVC_ERROR_INTERMITTENT + 130
    userMessage = "The dataset creation failed because it already exists. Please refresh the page."
    devMessage = "Unexpected error creating a dataset, possibly caused by concurrent actions."

    def __init__(self, exception=None):
        super().__init__(exception=exception)


class IntermittentProjectTemplateUnavailable(ServiceError):
    """The reference template for the project is currently unavailable.

    It may be a temporary issue in accessing the remote template, or it may have been deleted,
    moved, or otherwise not-accessible.
    """

    code = SVC_ERROR_INTERMITTENT + 140
    userMessage = (
        "The reference template for the project is currently unavailable."
        " It may be a temporary problem, or the template may not be accessible anymore."
    )
    devMessage = (
        "Error accessing the project template. This may be temporary, or the project may not be accessible anymore."
    )

    def __init__(self, exception=None):
        super().__init__(exception=exception)


class IntermittentWorkflowNotFound(ServiceError):
    """An operation failed because a workflow could not be found.

    It may be a synchronization error happening when two or more concurrent operations overlap
    and one tries to read content not yet created.
    """

    code = SVC_ERROR_INTERMITTENT + 150
    userMessage = "The workflow '{name_or_id}' could not be found. Check that the name/id is correct and try again."
    devMessage = "Unexpected error on workflow '{name_or_id}', possibly caused by concurrent actions."

    def __init__(self, exception=None, name_or_id=ERROR_NOT_AVAILABLE):
        super().__init__(
            userMessage=self.userMessage.format(name_or_id=name_or_id),
            devMessage=self.devMessage.format(name_or_id=name_or_id),
            exception=exception,
        )


class IntermittentTimeoutError(ServiceError):
    """An operation timed out."""

    code = SVC_ERROR_INTERMITTENT + 200
    userMessage = "The operation was taking too long. Please try it again."
    devMessage = "Timeout error. See Sentry exceptions for details."

    def __init__(self, exception=None):
        super().__init__(exception=exception)


class IntermittentLockError(ServiceError):
    """An operation was deadlocked."""

    code = SVC_ERROR_INTERMITTENT + 201
    userMessage = "The operation was taking too long. Please try it again."
    devMessage = "Deadlocked operation. See Sentry exceptions for details."

    def __init__(self, exception=None, message=ERROR_NOT_AVAILABLE):
        super().__init__(exception=exception)


class IntermittentRedisError(ServiceError):
    """An error occurred when interacting with Redis."""

    code = SVC_ERROR_INTERMITTENT + 202
    userMessage = "The servers could not run the request operation. Please try it again."
    devMessage = "Redis error. See Sentry exceptions for details."

    def __init__(self, exception=None):
        super().__init__(exception=exception)


class IntermittentCacheError(ServiceError):
    """An operation in the cache failed."""

    code = SVC_ERROR_INTERMITTENT + 203
    userMessage = "A server-side operation unexpectedly failed. Please try again."
    devMessage = "Cache error. See Sentry exceptions for details."

    def __init__(self, exception=None):
        super().__init__(exception=exception)
